"""Fast Research Interface (FRI) Server to remotely control Kuka LBR IV robot.

This module provides a server and client classes of Kuka's 
Fast Research Interface (FRI). 
Provided classes:
FRIServer: send control commands to and get status from KUKA LWR IV via UDP/IP.
FRIClient: receive commands and send status via UDP/IP of emulated FRI client.  
"""

import struct
import time

from fripy.friemu import FRIClientBase
import fripy.fricomm as fc
from fripy.fricomm import FRIMode, FRIQuality


class FRIClient(FRIClientBase):
    
    """Emulate the client side of the FRI interface. 
    Provided methods:
    open(dt) open FRI in monitor mode with sampling period dt miliseconds.
    start() start FRI in command mode.
    stop() stop FRI and switch back to monitor mode.
    close() close FRI, shut down network connection.  
    set_control_strategy() set the KRL/FRI control $strategy variable.
    send_data() send the msr structure to the server.
    recv_data() receive and validate the cmd structure from the server.
    """

    def __init__(self, verbose=True, debug=False):
        """Extends base class with the FRI methos: open, start, stop, close.
    
        By default, the client IP address is 127.0.0.1 and port 49938.
        Input Arguments:
        verbose=True, print a message anytime the interface's state changes.
        """
        
        FRIClientBase.__init__(self)     
        self._verbose = verbose
        self._debug = debug
        self._quality = FRIQuality.BAD
        self._mode = FRIMode.OFF
        self._control = fc.FRI_CTRL_OTHER
        
        # Timing
        self._packet_time = 0
        
        # Packet sequence
        self._local_packet_sequence = 0
        self._remote_packet_sequence = 0
        
        # Connection information
        self._latency = 0
        self._jitter = 0
        
        # Quality measurement
        self._PACKET_THRESHOLD = 2
        self._valid_packet_count = 0
        self._overall_valid_packet_count = 0
        self._overall_invalid_packet_count = 0
        
        self._dont_answer = False
        
    def start(self):
        """Emulate FRISTART, switch to command mode if quality is good."""
        if self._mode == FRIMode.CMD:
            return
        if self._quality >= FRIQuality.OK:
            self._switch_mode(FRIMode.CMD, "FRISTART")
        else:
            print("ERROR: Cannot enter CMD mode. Connection quality too low.")
       
    def stop(self):
        """Emulate FRISTOP, switch to monitor mode. """
        if self._mode == FRIMode.CMD:
            self._switch_mode(FRIMode.MON, "FRISTOP")

    def open(self, dt):
        """Emulate FRIOPEN, open the interface with a given sampling period.
        
        Input Arguments:
        dt: the sampling period in seconds (a positive floating point number).
        The interface state is set to monitor mode. """
        if (self._mode is not FRIMode.OFF):
            raise self.OpenError('Interface already open.')

        self.dt = dt
        if self._verbose:
            print("Setting sample time to: {0:.4f}s".format(self.dt))
        self.sock.settimeout(self.dt)
        self._mode = FRIMode.MON
        self._quality = FRIQuality.BAD
        self.TIMEFRAME = dt
        if self._verbose:
            print("Connection opened")
        
    def close(self):
        """Emulate FRICLOSE, close network connection."""
        if self._mode == FRIMode.OFF:
            return
        self._mode = FRIMode.OFF
        self.sock.close()
        if self._verbose:
            print("Connection closed")
    
    def set_control_strategy(self, control):
        """Emulate setting the value of the KRL variable $strategy.
        
        Input Arguments:
        control: any of the following constants found in fricomm module:
        FRI_CTRL_POSITION, for joint position control $strategy=10, 
        FRI_CTRL_JNT_IMP, for joint impedance control $strategy=30,
        FRI_CTRL_CART_IMP, for cartesian impedance control $strategy=20.
        """
        if self._mode == FRIMode.CMD:
            print("Cannot change control strategy in CMD mode!")
        elif self._control != control:
            if control == fc.FRI_CTRL_OTHER:
                self._control = fc.FRI_CTRL_OTHER
                print("Set control strategy to Other.")
            if control == fc.FRI_CTRL_POSITION:
                self._control = fc.FRI_CTRL_POSITION
                print("Set control strategy to Joint Position Control.")
            if control == fc.FRI_CTRL_CART_IMP:
                self._control = fc.FRI_CTRL_CART_IMP
                print("Set control strategy to Cartesian Impedance Control.")
            if control == fc.FRI_CTRL_JNT_IMP:
                self._control = fc.FRI_CTRL_JNT_IMP
                print("Set control strategy to Joint Impedance Control.")
    
    def send_data(self):
        """Fill the msr.head/intf and send the msr structure to the server."""
        now = time.time()
        
        # HEADER     
        self._local_packet_sequence += 1 # TODO: Increase it here?
        self.msr.head.sendSeqCount = self._local_packet_sequence
        self.msr.head.reflSeqCount = self._remote_packet_sequence
        self.msr.head.packetSize = struct.calcsize(self.msr._fmt) 
        self.msr.head.datagramId = fc.FRI_DATAGRAM_ID_MSR
        
        # FRI STATE
        self.msr.intf.timestamp = now
        self._packet_time = now
        self.msr.intf.state = FRIMode.toPacketValue(self._mode)
        self.msr.intf.quality = FRIQuality.toPacketValue(self._quality)
        self.msr.intf.desiredMsrSampleTime = self.TIMEFRAME
        self.msr.intf.desiredMsrSampleTime = self.TIMEFRAME
        self.msr.intf.safetyLimits = 1.0
        
        # ROBOT
        self.msr.robot.control = self._control
        
        # STATISTICS
        self.msr.intf.stat.answerRate = self._overall_valid_packet_count / self._overall_invalid_packet_count if self._overall_invalid_packet_count > 0 else 1
        self.msr.intf.stat.latency = self._latency
        self.msr.intf.stat.jitter = self._jitter
        self.msr.intf.stat.missRate = self._overall_invalid_packet_count / self._overall_valid_packet_count if self._overall_valid_packet_count > 0 else 0
        self.msr.intf.stat.missCounter = self._overall_invalid_packet_count

        self.send_msr()
         
    def recv_data(self):
        """Receive the cmd structure. Decrease FRI quality if invalid cmd."""
        # FIXME: deep nesting UGLY!!!!!
        if self._mode == FRIMode.MON or self._mode == FRIMode.CMD:
            if self._debug:
                    print("{0:.4f}s: recv_data in mode {1}".format(time.time(), FRIMode.toString(self._mode)))
            if self.recv_cmd():
                if self._debug:
                        print("{0:.4f}s: recv_cmd returned True".format(time.time()))
                if self._validate_sequence():
                    if self._debug:
                            print("{0:.4f}s: _validate_sequence returned True".format(time.time()))
                    if self._validate_connection():
                        if self._debug:
                                print("{0:.4f}s:_validate_connection returned True".format(time.time()))
                        if self._validate_packet_info():
                            if self._debug:
                                    print("{0:.4f}s: _validate_packet_info returned True, calling _register_valid_packet".format(time.time()))
                            self._register_valid_packet()
            else:
                self._decrease_quality('Packet timed out')
            if time.time() - self._packet_time > self.TIMEFRAME:
                if self._debug:
                    print("{0:.4f}s: time.time() - self._packet_time > self.TIMEFRAME".format(time.time()))
                self.send_data()
      
    def _decrease_quality(self, reason):
        """Decrease connection quality and switch to MON if necessary."""
        if self._quality == FRIQuality.BAD:
            return
        self._quality = max(FRIQuality.BAD, self._quality - 1)
        if self._mode == FRIMode.CMD:
            self._switch_mode(FRIMode.MON, "Connection quality decreased")
        if self._verbose:
            print("Decreased quality to %s (%s)" % (FRIQuality.toString(self._quality), reason))
        
    def _increase_quality(self):
        """Increase connection quality."""
        if self._quality == FRIQuality.PERFECT:
            return
        self._quality = min(FRIQuality.PERFECT, self._quality + 1)
        if self._verbose:
            print("Increased quality to %s" % (FRIQuality.toString(self._quality),))
        
    def _switch_mode(self, mode, reason):
        """Switch FRI state."""
        self._mode = mode
        if self._verbose:
            print("Switched mode to %s (%s)" % (FRIMode.toString(self._mode), reason))

    def _register_valid_packet(self):
        """Called if a valid packet has arrived."""
        self._valid_packet_count += 1
        self._overall_valid_packet_count += 1
        if self._valid_packet_count >= self._PACKET_THRESHOLD:
            self._valid_packet_count = 0
            self._increase_quality()
        
    def _validate_sequence(self):
        """Detects missing packets in the packet stream."""
        expectedPacketSequence = self._remote_packet_sequence + 1
        # TODO: Reset the package sequence here??
        self._remote_packet_sequence = self.cmd.head.sendSeqCount 
        distance = self.cmd.head.sendSeqCount - expectedPacketSequence
        if not distance == 0:
            self._valid_packet_count = 0
            self._overall_invalid_packet_count += 1
            if distance > 0:
                self._decrease_quality("Remote packet sequence is %d ahead" % (distance,))
            elif distance < 0:
                self._decrease_quality("Remote packet sequence is %d behind" % (-distance,))
            return False
        return True
                
    def _validate_connection(self):
        """Detects obsolete packets."""
        distance = self.cmd.head.reflSeqCount - self._local_packet_sequence
        
        if not distance == 0:
            self._overall_invalid_packet_count += 1
            if distance > 0:
                self._decrease_quality("Reflected packet sequence is %d ahead" % (distance,))
            else:
                self._decrease_quality("Reflected packet sequence is %d behind" % (-distance,))
                self._dont_answer = True # An old package has been captured. Leave it unanswered!
            return False
        else:
            previousLatency = self._latency
            self._latency = time.time() - self._packet_time
            self._jitter = previousLatency - self._latency
            return True
    
    def _validate_packet_info(self):
        """Makes sure a packet has the right id and size."""
        if not struct.calcsize(self.cmd._fmt) == self.cmd.head.packetSize:
            self._decrease_quality("Invalid packet size")
            self._overall_invalid_packet_count += 1
            return False
        elif not self.cmd.head.datagramId == fc.FRI_DATAGRAM_ID_CMD:
            self._decrease_quality("Invalid datagram id")
            self._overall_invalid_packet_count += 1
            return False
        return True

