"""Create Python classes corresponding to C structures.

Classes provided:
CStruct: defines and keeps track of Python/C structures. 
CStruct methods:
def_new_pystruct(str) creates a new class from a string describing a C struct.
"""
import struct

from fripy.fricomm import * # constants defining size of arrays
import time

class _PyStruct(object):

    """It is the base class on which new PyStructs are defined according to 
    the C struct definition. This class provided general methods, but has no
    attributes, they are added dynamically to the subclasses. An instance
    of the subclass is automatically created and initialised
    with predefined values."""

    def __init__(self, fmt, names, classes, initargs):
        """Create a subclass with the given input parameters. The subclass only
        adds new properties named 'names' of class 'classes' and makes an 
        instance initialised with 'initargs'. Defines the types format 'fmt'
        """
        self._fmt = fmt
        self._alldatalen = len(fmt)
        self._alldata = [None] * self._alldatalen
        self._names = names 
        self._datalen = len(names)
        self._data = [None] * self._datalen
        self._datatype = type(self._data)
        self._classes = classes
        self._initargs = initargs
        def make_prop(k):
            def _set(self, val):
                self._data[k] = val
            def _get(self):
                return self._data[k]
            return property(_get, _set)
        # Add new attributes to the PyStruct (sub)class and instance (self).
        for k in range(self._datalen):
            setattr(type(self), self._names[k], make_prop(k))
            self.__setattr__(self._names[k], 
                             self._classes[k](*self._initargs[k]))
    
    def _set_raw(self, input):
        if not isinstance(input, type(self._alldata)):
            print('Error: wrong input type. Input argument must be', 
                  type(self._alldata))
            return
        if len(input) != self._alldatalen:
            print('Error: wrong input length. Given:',
                   len(input), '. Required:', self._alldatalen)
            return
        for x in input:
            try:
                float(x)
            except (ValueError, TypeError):
                print('Error: the initialization data must be numeric.',
                 'Given', x, 'in pos' )
                return
        self._alldata = input
        
    
    # Works iteratively in the structures tree, going down until it finds
    # a non-PyStruct object (i.e. an actual value, such as an int). When 
    # found, it either copies the value in _alldata[pos] (if no input_data)
    # or sets the corresponding attribute with the given input_data[pos]. 
    def get_int(self, obj, obj_name, parent, input_data):
        if isinstance(obj, _PyStruct):
            for child_name in obj._names:
                child = obj.__getattribute__(child_name)
                self.get_int(child, child_name, obj, input_data)
        else:
            if input_data is not None:
                if isinstance(obj, type([])):
                    datalist = []
                    for element in obj:
                        datalist.append(self._alldata[self._pos]) 
                        self._pos += 1
                    parent.__setattr__(obj_name, datalist)
                    self._pos -= 1  # FIXME UGLY: accounts for += 1 outside
                else:                    
                    parent.__setattr__(obj_name, self._alldata[self._pos])
            else:
                if isinstance(obj, type([])):
                    for element in obj:
                        self._alldata[self._pos] = element 
                        self._pos += 1
                    self._pos -= 1  # FIXME UGLY: accounts for += 1 outside
                else:
                    self._alldata[self._pos] = obj
                
            self._pos += 1
            
    def _update_alldata(self, input_data=None):
        #print "def get int {0:.4f}".format(time.time())
        self._pos = 0
        if input_data is not None:
            self._set_raw(input_data)
            #print "set raw {0:.4f}".format(time.time())
        self.get_int(self, None, None, input_data)
        
        #print "getint {0:.4f}".format(time.time())
        
    def get_data(self):
        self._update_alldata()
        return self._alldata
    
    def set_data(self, data):
        data = list(data)
        #print "tolist {0:.4f}".format(time.time())
        self._update_alldata(data)
        #print "updateall {0:.4f}".format(time.time())
        

class CStruct(object):

    """Create Python classes that resemble C structs. 
    
    The data can be accessed in Python similarly as in C. For example, using
    an expression like mainstruct.substruct.array[0] = 1. substruct must be 
    defined before it can be used in mainstruct."""
    
    def __init__(self):
        """Define basic C types and its corresponding Python representation."""
        self._pystructs = dict()
        self._pytypes = dict(fri_uint16_t='H', fri_uint32_t='I', 
                           fri_int32_t='i', fri_float_t='f')
        self._classes = dict(fri_uint16_t=int, fri_uint32_t=int, 
                                 fri_int32_t=int, fri_float_t=float)
        self._initargs = dict(fri_uint16_t=[0], fri_uint32_t=[0], 
                               fri_int32_t=[0], fri_float_t=[0.])
        self._arraylens = dict(LBR_MNJ=7, FRI_USER_SIZE=16,
                               FRI_CART_FRM_DIM=12, FRI_CART_VEC =  6)

    def _get_pystruct_data_from_str(self, struct_str):
        names = []
        fmts = []
        classes = []
        inits = []
        struct_lines = struct_str.splitlines()
        typedef = struct_lines.pop()
        for line in struct_lines:
            try:
                ctype, name, arraylenstr = line.split()
                arraylen = eval(arraylenstr)
            except ValueError:
                ctype, name = line.split()
                arraylen = 1
            names.append(name)
            fmts.append(self._pytypes[ctype] * arraylen)
            if arraylen == 1:
                classes.append(self._classes[ctype])
                inits.append(self._initargs[ctype])
            else:
                classes.append(type([]))
                inits.append([self._initargs[ctype] * arraylen])
        fmtstr = ''.join(fmts)
        return [typedef, names, fmtstr, classes, inits]

    def _add_new_PyStruct_def_to_self(self, typedef, fmtstr, obj, cls, init):
        self.__setattr__(typedef, obj)
        self._pytypes.setdefault(typedef, fmtstr)
        self._pystructs.setdefault(typedef, obj)
        self._classes.setdefault(typedef, cls)
        self._initargs.setdefault(typedef, init)
        
    def _create_new_PyStruct_subclass(self, typedef, names, fmtstr, classes, 
                                      initargs):
        base = _PyStruct
        cls = type(typedef, (base,), {})
        init = [fmtstr, names, classes, initargs]
        obj = cls(*init)
        self._add_new_PyStruct_def_to_self(typedef, fmtstr, obj, cls, init)
        
    def def_new_pystruct(self, struct_str):
        """Define a new pystruct based on a input string describing a C struct.
        
        The input string format is as follows:
        each structure field should be in its own line, no spaces/tabs before
        or after. Arguments are separated by spaces. First argument is the 
        C type, second argument is the variable name, and the third (optional)
        is the array size (it can be an expression). Last line is the name
        of the structure.  
        
        Example:
        fri_int32_t float1
        fri_int32_t intarray1 LBR_MNJ
        fri_float_t float2Darray LBR_MNJ*FRI_CART_VEC
        tFriHeader another_structure
        tFriTestStructName
        """
        pystruct_data = self._get_pystruct_data_from_str(struct_str)
        self._create_new_PyStruct_subclass(*pystruct_data)
    
