'''Emulate the Fast Research Interface (FRI).

This module emulates Kuka's Fast Research Interface (FRI). 
Provided classes:
FRIEmu: Aims to emulate the KRL FRI client (LWR side) as complete as possible.
FRIEmuServer: send and receive the tFriCmdData and tFriMsrData, respectively. 
'''

import socket
import struct
import threading

from fripy import cstruct
from fripy import fricomm as fc

import sys, time

class _FRICommBase(object):

    """Provide base data and socket methods for FRI client and server. """
    
    def __init__(self, IP_ADDR_PORT=("127.0.0.1", 49938)):
        """Create tFriMsrData, tFriCmdData structs, and set UDP/IP addresses."""
        cs = cstruct.CStruct()
        cs.def_new_pystruct(fc.tFriHeader)
        cs.def_new_pystruct(fc.tFriKrlData)
        cs.def_new_pystruct(fc.tFriIntfStatistics)
        cs.def_new_pystruct(fc.tFriIntfState)
        cs.def_new_pystruct(fc.tFriRobotState)
        cs.def_new_pystruct(fc.tFriRobotData)
        cs.def_new_pystruct(fc.tFriMsrData)
        cs.def_new_pystruct(fc.tFriRobotCommand)
        cs.def_new_pystruct(fc.tFriCmdData)
        self.msr = cs.tFriMsrData
        self.cmd = cs.tFriCmdData
        self.IP_ADDR_PORT = IP_ADDR_PORT
        self.peer_addr = ()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._verbose = False
        self._debug = False
        
    def _recv_udp(self, fmt):
        bytes_ = struct.calcsize(fmt)
        
        if self._debug:
            starttime = time.time()
                    
        bindat, self.peer_addr = self.sock.recvfrom(bytes_)
        
        if self._debug:
            print(("{0:.4f}s: DATA RECEIVED from {1}, "
                   "length: {2}, took {3:.4f}s"
                   ).format(time.time(), self.peer_addr, len(bindat), 
                                           time.time()-starttime))
        #", DATA:-->{4}<--", bindat
        #if self._debug:
        #    starttime = time.time()
        
        dat = struct.unpack(fmt, bindat) 
        
        #if self._debug:
        #    print("{0:.4f}s: DATA UNPACK took {1:.4f}".format(time.time(), time.time()-starttime))
        
        return dat

    def _send_udp(self, fmt, dat):
        bindat = struct.pack(fmt, *dat)
        
        if self._debug:
            starttime = time.time()
        if self.peer_addr == ():
            print(("{0:.4f}s: Warning: DATA SEND canceled,"
                   " because remote peer not known yet.").format(time.time()))
            return False
                
        bytes_ = self.sock.sendto(bindat, self.peer_addr)
        
        if self._debug:
            print(("{0:.4f}s: DATA SEND to {1}, "
                  "length: {2}, took {3:.4f}s").format(time.time(), 
                                          self.peer_addr, bytes_, 
                                          time.time()-starttime))
        #", DATA:-->{4}<--", bindat
        
    class _FRIError(Exception):
        """Base class for FRI internal exceptions."""
        pass


class FRIClientBase(_FRICommBase):
    
    """Emulate the KRL FRI client internal functionality.
    
    Extends _FRICommBase with UDP communication specific to the client.
     Available methods: 
     send_msr() send msr, an instance of the tFriMsrData structure.
     recv_cmd() receive cmd, an instance of the tFriCmdData structure.
     """
    
    def __init__(self, IP_ADDR_PORT=("127.0.0.1", 49938)):
        """Define communication parameters.
        
        Extends the superclass method with a non-binding UDP socket."""
        _FRICommBase.__init__(self, IP_ADDR_PORT)
        self.peer_addr = self.IP_ADDR_PORT
        
    def send_msr(self):
        """Send msr, an instance of the tFriMsrData structure, via UDP."""
        
        if self._debug:
            print(("{0:.4f}s: SEND MSR: sendSeqCount: {1}, reflSeqCount: {2},"
                   " packetSize: {3}, datagramId: {4}").format(time.time(), 
                                                               self.msr.head.sendSeqCount, 
                                                               self.msr.head.reflSeqCount, 
                                                               self.msr.head.packetSize, 
                                                               self.msr.head.datagramId))
        data = self.msr.get_data()
        fmt = self.msr._fmt
        self._send_udp(fmt, data)
        return False
        
    def recv_cmd(self):
        """Receive cmd, an instance of the tFriCmdData structure, via UDP."""
        ontime = False
        fmt = self.cmd._fmt
        try:
            dat = self._recv_udp(fmt)
            ontime = True
            
            if self._debug:
                starttime = time.time()
                
            self.cmd.set_data(dat)
            
            if self._debug:
                print(("{0:.4f}s: CMD SET DATA took "
                      "{1:.4f}s").format(time.time(), time.time()-starttime))
            if self._debug:
                print(("{0:.4f}s: RECV CMD: sendSeqCount: {1}, reflSeqCount: {2},"
                       " packetSize: {3}, datagramId: {4}").format(time.time(), 
                                                                   self.cmd.head.sendSeqCount, 
                                                                   self.cmd.head.reflSeqCount, 
                                                                   self.cmd.head.packetSize, 
                                                                   self.cmd.head.datagramId))
                            
        except (socket.timeout, ):
            print(("{0:.4f}s: Error: SocketTimeout ({1}s) at "
                   "receive: {2}").format(time.time(), 
                                          self.sock.gettimeout(), 
                                          sys.exc_info()[0]))
        except:
            print(("{0:.4f}s: Unexpected error at "
                  "receive: {1}").format(time.time(), sys.exc_info()[0]))
            raise
        
        return ontime
            
    class OpenError(_FRICommBase._FRIError):
        pass
    


class FRIServerBase(_FRICommBase):
    
    """
    Provide basic FRI server communication functionality. 
    
    Extends _FRICommBase with UDP communication specific to the server.
    It does not process the received data or the data to be sent.
    
    send_cmd() send cmd, an instance of the tFriCmdData structure.
    recv_msr() receive msr, an instance of the tFriMsrData structure.
    """
    
    def __init__(self, IP_ADDR_PORT=("127.0.0.1", 49938)):
        """Define communication parameters and binds to the given IP address.
        
        Extends the superclass method with a binding UDP socket.
        Arguments:
        IP_ADDR_PORT: a tuple containing the IP address and port. If not 
        specified, defaults to ("127.0.0.1", 49938)
        """
        _FRICommBase.__init__(self, IP_ADDR_PORT)
        
        self.sock.bind(self.IP_ADDR_PORT)
        
    def send_cmd(self):
        """Send cmd, an instance of the tFriCmdData structure, via UDP."""
        if self._debug:
            print(("{0:.4f}s: SEND CMD: sendSeqCount: {1}, reflSeqCount: {2},"
                   " packetSize: {3}, datagramId: {4}").format(time.time(), 
                                                               self.cmd.head.sendSeqCount, 
                                                               self.cmd.head.reflSeqCount, 
                                                               self.cmd.head.packetSize, 
                                                               self.cmd.head.datagramId))
        data = self.cmd.get_data()
        fmt = self.cmd._fmt
        self._send_udp(fmt, data)
        
    def recv_msr(self):
        """Receive msr, an instance of the tFriMsrData structure, via UDP."""
        fmt = self.msr._fmt
        
        ontime = False
        
        try:
            dat = self._recv_udp(fmt)
            ontime = True
            
            if self._debug:
                starttime = time.time()
                
            self.msr.set_data(dat)
            
            if self._debug:
                print(("{0:.4f}s: MSR SET DATA took "
                       "{1:.4f}s").format(time.time(), time.time()-starttime))            
            
            if self._debug:
                print(("{0:.4f}s: RECV MSR: sendSeqCount: {1}, reflSeqCount: {2},"
                       " packetSize: {3}, datagramId: {4}").format(time.time(), 
                                                                 self.msr.head.sendSeqCount, 
                                                                 self.msr.head.reflSeqCount, 
                                                                 self.msr.head.packetSize, 
                                                                 self.msr.head.datagramId))
        
        except (socket.timeout, ):
            print(("{0:.4f}s: Error: SocketTimeout ({1}s) at "
                   "receive: {2}").format(time.time(), self.sock.gettimeout(), sys.exc_info()[0]))
        except:
            print(("{0:.4f}s: Unexpected error at "
                   "receive: {1}").format(time.time(), sys.exc_info()[0]))
            raise
        
        return ontime
    
    def _dataExchange(self):
        while True:
            self.recv_msr()
            self.send_cmd()
    
    def background_recv_send(self):
        """Continuously execute recv_msr and send_cmd methods in a thread."""
        self.thr = threading.Thread(target=self._dataExchange)
        self.thr.start()
