"""Fast Research Interface (FRI) Server to remotely control Kuka LBR IV robot.

This module provides a server and client classes of Kuka's 
Fast Research Interface (FRI). 
Provided classes:
FRIServer: send control commands to and get status from KUKA LWR IV via UDP/IP.
FRIClient: receive commands and send status via UDP/IP of emulated FRI client.  
"""

import struct
import time

from fripy.friemu import FRIServerBase
import fripy.fricomm as fc
from fripy.fricomm import LBR_MNJ

class FRIServer(FRIServerBase):

    """Remotely control Kuka LBR IV through FRI via UDP/IP. 
    
    Provided methods:
    recv_data(): receive the msr structure via UDP/IP.
    send_data(): fill in the cmd.head and send the cmd structure via UDP/IP.
    ctl_jnt_pos(): joint position controller.
    ctl_jnt_imp(): joint impedance controller.
    ctl_cts_imp(): cartesian impedance controller.
    print_fri_statistics(): display FRI network performance statistics.
    print_fri_status(): display interface status.
    Instance variables:
    msr: the FRI msr structure.
    cmd: the FRI cmd structure.
    """

    def __init__(self, IP_ADDR_PORT=None, simulation=False, timeout=None, 
                 verbose=True, debug=False):
        """Extends base class with controller and server specific methods.
        
        Extends base class with server controllers methods for the three LBR 
        control strategies ($strategy=10, 20 and 30), and with server specific 
        display and communication methods.
        Input arguments:
        IP_ADDR_PORT=None, a tuple of the form ('127.0.0.1', 49938).
        simulation=False, if a connection is to the real robot or simulation.
        timeout=None, puts the socket receive in blocking mode. Set timeout
        to a positive float to detect packets arriving timeout seconds late.   
        If simulation=True, IP_ADDR_PORT=('127.0.0.1', 49938),
        if simulation=False and IP_ADDR_PORT=None, then 
        IP_ADDR_PORT=('192.168.0.2', 49938), 
        if simulation=False and IP_ADDR_PORT is not None, then 
        IP_ADDR_PORT is the given IP address and port tuple.
        """
        if simulation is True:
            IP_ADDR_PORT = ('127.0.0.1', 49938)
        elif (IP_ADDR_PORT is None):
            IP_ADDR_PORT = ('192.168.0.2', 49938)
        print('IP address and port number:', IP_ADDR_PORT)
        FRIServerBase.__init__(self, IP_ADDR_PORT)
        self.sock.settimeout(timeout)
        self.seqCount = 0
        self.sizeof_tFriCmdData = struct.calcsize(self.cmd._fmt)
        self._verbose = verbose
        self._debug = debug
    
    def recv_data(self):
        """Receive the msr structure via UDP/IP."""
        return self.recv_msr()
    
    def send_data(self):
        """Fill in the cmd.head and send the cmd structure via UDP/IP. """
        self.seqCount += 1
        USHRT_MAX = 65535
        if self.seqCount >= USHRT_MAX:
            self.seqCount = 0
        
        self.cmd.head.sendSeqCount = self.seqCount
        self.cmd.head.reflSeqCount = self.msr.head.sendSeqCount
        self.cmd.head.datagramId = fc.FRI_DATAGRAM_ID_CMD
        self.cmd.head.packetSize =  self.sizeof_tFriCmdData
        self.send_cmd()
        # FIXME: return value?
    
    def ctl_jnt_pos(self, newJntPos, sendData=True):
        """Control the angular position of the robot joints ($strategy = 10).
        
        If FRI is not in command mode, mirror the current measured position. 
        If FRI is in command mode, send the desired absolute joint position. 
        Input arguments:
        newJntPos is a list of size LBR_MNJ, with the desired absolute position
        of each joint in radians.
        sendData=True, send data via UDP/IP or not.  
        """
        self.cmd.cmd.cmdFlags = fc.FRI_CMD_JNTPOS
        if ((self.msr.intf.state is not fc.FRI_STATE_CMD) or
            (bool(self.msr.robot.power) is False)):
            for i in range(LBR_MNJ):
                self.cmd.cmd.jntPos[i] = (self.msr.data.cmdJntPos[i] + 
                                          self.msr.data.cmdJntPosFriOffset[i])
        else:
            for i in range(LBR_MNJ):
                self.cmd.cmd.jntPos[i] = newJntPos[i]
        
        if (sendData is True):
            self.send_data()
            
    def ctl_jnt_imp(self, newJntPos=None, newJntStf=None, 
                    newJntDmp=None, newJntTrq=None, sendData=True):
        """Control torque of the robot joints (joint impedance, $strategy = 30).
        
        If FRI is not in command mode, mirror the current measured position. 
        If FRI is in command mode, send the desired absolute joint position. 
        Input arguments:
        newJntPos=None, a list of size LBR_MNJ, not sure exactly what it does. 
        newJntStf=None, a list of size LBR_MNJ, with the stiffness of each 
        individual joint. Set all elements to zero for no stiffness, otherwise 
        not sure exactly what it does.
        newJntDmp=None, a list of size LBR_MNJ, with the damping of each 
        individual joint. Set all elements to zero for no damping, otherwise 
        not sure exactly what it does.
        newJntTrq=None, a list of size LBR_MNJ, with the torque to be added
        to the torque computed by the KRL internal controller. If newJntPos=None
        and newJntStf=newJntDmp=[0]*LBR_MNJ, the KRL controller continuously 
        computes the gravity torque only, and newJntTrq is to compensate for 
        all other effects (inertia, friction, etc.).
        sendData=True, send data via UDP/IP or not.        
        """
        
        self.cmd.cmd.cmdFlags = 0
        if (newJntPos is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_JNTPOS
            if ((self.msr.intf.state is not fc.FRI_STATE_CMD) or
                (bool(self.msr.robot.power) is False)):
                for i in range(LBR_MNJ):
                    self.cmd.cmd.jntPos[i] = (self.msr.data.cmdJntPos[i] + 
                                          self.msr.data.cmdJntPosFriOffset[i])
            else:
                for i in range(LBR_MNJ):
                    self.cmd.cmd.jntPos[i] = newJntPos[i]
        if (newJntStf is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_JNTSTIFF
            for i in range(LBR_MNJ):
                    self.cmd.cmd.jntStiffness[i] = newJntStf[i]
        if (newJntDmp is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_JNTDAMP
            for i in range(LBR_MNJ):
                    self.cmd.cmd.jntDamping[i] = newJntDmp[i]
        if (newJntTrq is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_JNTTRQ
            for i in range(LBR_MNJ):
                    self.cmd.cmd.addJntTrq[i] = newJntTrq[i]
        if (sendData is True):
            self.send_data()

    def ctl_cts_imp(self, newCtsPos=None, newCtsStf=None, newCtsDmp=None,
                    newTcpFT=None, newJntNS=None, sendData=True):
        """Control cartesian position/orientation of end effector ($strategy=20)

        FIXME: This documentation is not up-to-date. Function not fully tested.    
        newCtsPos=None, a list of size FRI_CART_FRM_DIM, not sure what it does. 
        newCtsStf=None, a list of size FRI_CART_VEC, with the stiffness of each 
        cartesian direction. Set all elements to zero for no stiffness, 
        otherwise not sure exactly what it does.
        newCtsDmp=None, a list of size FRI_CART_VEC, with the damping of each 
        cartesian direction. Set all elements to zero for no damping, otherwise 
        not sure exactly what it does.
        newTcpFT=None, a list of size FRI_CART_VEC, not sure what it does.
        newJntNS=None, a list of size LBR_MNJ, not sure what it does.
        sendData=True, send data via UDP/IP or not.        
        
    """
        self.cmd.cmd.cmdFlags = 0
        if (newCtsPos is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_CARTPOS
            for i in range(fc.FRI_CART_FRM_DIM):
                    self.cmd.cmd.cartPos[i] = newCtsPos[i]
        if (newCtsStf is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_CARTSTIFF
            for i in range(fc.FRI_CART_VEC):
                    self.cmd.cmd.cartStiffness[i] = newCtsStf[i]
        if (newCtsDmp is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_CARTDAMP
            for i in range(fc.FRI_CART_VEC):
                    self.cmd.cmd.cartDamping[i] = newCtsDmp[i]
        if (newTcpFT is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_TCPFT
            for i in range(fc.FRI_CART_VEC):
                    self.cmd.cmd.addTcpFT[i] = newTcpFT[i]
        if (newJntNS is not None):
            self.cmd.cmd.cmdFlags |= fc.FRI_CMD_JNTPOS
            if ((self.msr.intf.state is not fc.FRI_STATE_CMD) or
                (bool(self.msr.robot.power) is False)):
                # FIXME: KUKA's sample code does not do this 'if' check (why?)
                for i in range(LBR_MNJ):
                    self.cmd.cmd.jntPos[i] = (self.msr.data.cmdJntPos[i] + 
                                          self.msr.data.cmdJntPosFriOffset[i])
            else:
                for i in range(LBR_MNJ):
                    self.cmd.cmd.jntPos[i] = newJntNS[i]
        if (sendData is True):
            self.send_data()            
        
    def print_fri_status(self):
        """Print FRI interface status information."""
        print('Interface Status')
        print(' - time stamp:', self.msr.intf.timestamp)
        print(' - state:', self.msr.intf.state)
        print(' - quality:', self.msr.intf.quality)
        print(' - desired msr sample time:', self.msr.intf.desiredMsrSampleTime)
        print(' - desired cmd sample time:', self.msr.intf.desiredCmdSampleTime)
        print(' - safety limits:', self.msr.intf.safetyLimits)
    
    def print_fri_statistics(self):
        """Print FRI network performance statistics. """
        print('Interface Statistic')
        print(' - answer rate:', self.msr.intf.stat.answerRate)
        print(' - latency:', self.msr.intf.stat.latency)
        print(' - jitter:', self.msr.intf.stat.jitter)
        print(' - miss rate:', self.msr.intf.stat.missRate)
        print(' - miss counter:', self.msr.intf.stat.missCounter)
        
    def _exchange_data(self):
        self.send_data()
        self.recv_data()

