/*
 *   lwrmodels - Dynamic Robot Models of the KUKA Lightweight Robot IV
 *   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
 *
 *   This file is part of lwrmodels.
 *
 *   lwrmodels is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <model.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>


int invert_matrix(gsl_matrix * A, gsl_matrix * Ai);
int print_matrix(gsl_matrix * A, char * prefix);
int print_vector(gsl_vector * p, char * prefix);
int read_double_array(double * data, int count);

int main() {

    gsl_matrix * B = gsl_matrix_alloc(NUMJ_MOD,NUMJ_MOD);
    gsl_matrix * Binv = gsl_matrix_alloc(NUMJ_MOD,NUMJ_MOD);
    gsl_matrix * C = gsl_matrix_alloc(NUMJ_MOD,NUMJ_MOD);
    gsl_vector * g = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * tau_f = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * tau = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * tmp = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * q = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * dq = gsl_vector_alloc(NUMJ_MOD);
    gsl_vector * ddq = gsl_vector_alloc(NUMJ_MOD);
    
    int i;
    
    printf("Using a dynamic robot model for %i joints.\n", NUMJ_MOD);
    printf("These are the joints ");
    
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%i", SIMJOINTS[i]+1);
        
        if (i==NUMJ_MOD-2)
            printf(" and ");
        else if (i==NUMJ_MOD-1)
            printf(" ");
        else
            printf(", ");
    }
    printf("of a robot with %i joints.\n\n", NUMJ_ROB);
    
       
    
    printf("Enter %i joint positions (space separated): ", NUMJ_MOD);
    read_double_array(q->data, NUMJ_MOD); 
    printf("Enter %i joint velocities (space separated): ", NUMJ_MOD);
    read_double_array(dq->data, NUMJ_MOD);
    printf("Enter %i actuation torques (space separated): ", NUMJ_MOD); 
    read_double_array(tau->data, NUMJ_MOD);
    
    calc_gravity_torque_vector(g->data, q->data);
    calc_friction_torque_vector(tau_f->data, dq->data);
    calc_inertia_matrix((double (*)[NUMJ_MOD])(B->data), q->data);
    calc_coriolis_matrix((double (*)[NUMJ_MOD])(C->data), q->data, dq->data);
    
    printf("\n\nInput set to:\n");
    printf("      q = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", q->data[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T");
        else
            printf(", ");
    }
    printf("\n     dq = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", dq->data[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T");
        else
            printf(", ");
        
    }
    printf("\n    tau = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", tau->data[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T\n");
        else
            printf(", ");
        
    }
    printf("\n\n\nOutput:\n");
    
    
    print_matrix(B, "     B(q) = ");
    
    print_matrix(C, "  C(q,dq) = ");
    print_vector(g, "     g(q) = ");
    print_vector(tau_f, "tau_f(dq) = ");
    
    invert_matrix(B, Binv);
    print_matrix(Binv, "  B(q)^-1 = ");

    
    gsl_blas_dgemv(CblasNoTrans, 1.0, C, dq, 0.0, tmp);
    gsl_vector_sub(tau, tmp);
    gsl_vector_sub(tau, g);
    gsl_vector_sub(tau, tau_f);
    gsl_blas_dgemv(CblasNoTrans, 1.0, Binv, tau, 0.0, ddq);
    
    printf("Computed joint accelerations:\n");
    print_vector(ddq, "ddq = ");
 
    gsl_matrix_free(B);
    gsl_matrix_free(Binv);
    gsl_matrix_free(C);
    gsl_vector_free(g);
    gsl_vector_free(tau_f);
    gsl_vector_free(tau);
    gsl_vector_free(tmp);
    gsl_vector_free(q);
    gsl_vector_free(dq);
    gsl_vector_free(ddq);

    return 0;
}
    
/* Invert a matrix using GSL with LU factorization */
int invert_matrix(gsl_matrix * A, gsl_matrix * Ainv) {
    int signum;

    // create permutation matrix P
    gsl_permutation * P = gsl_permutation_alloc(A->size1);
    
    // make a copy of the original matrix
    gsl_matrix * ALU = gsl_matrix_alloc(A->size1, A->size2);
    gsl_matrix_memcpy(ALU, A);
    
    // perform LU-factorization and use result to get inverse
    gsl_linalg_LU_decomp(ALU, P, &signum);
    gsl_linalg_LU_invert(ALU, P, Ainv);
    
    gsl_matrix_free(ALU);
    gsl_permutation_free(P);
    
    return 0;
}

/* Print a matrix, print prefix in front of it */
int print_matrix(gsl_matrix * A, char * prefix) {
    int i=0, j=0, k=0, midline=0, mod2=0;
    
    char * prefix_space = (char *)malloc((strlen(prefix)+1)*sizeof(char));
    
    mod2 = A->size1 % 2;
    midline = (A->size1 - mod2)/2 ;
    
    k=0;
    while(*(prefix+k)!='\0') {
        *(prefix_space+k) = ' ';
        k++;       
    }
    *(prefix_space+k) = '\0';
    
    printf("%s--", prefix_space);
        
    for(j=0;j<A->size2-1;j++)
        printf("           ");
    printf("          --\n");
    
    for(i=0;i<A->size1;i++) {
        if(i==midline)
            printf("%s", prefix);
        else
            printf("%s", prefix_space);
    
        printf("| ");
        if(i==midline && mod2==0) {
            for(j=0;j<A->size2;j++) {
                    printf("          ");
                if (j<A->size2-1)
                    printf(" ");
                else
                    printf(" |\n");
            }
            printf("%s", prefix_space);
            printf("| ");
        }
        for(j=0;j<A->size2;j++) {
            printf("%10.3e", gsl_matrix_get(A,i,j));
            
            if (j<A->size2-1)
                printf(" ");
            else
                printf(" |\n");
        }
    }
    printf("%s", prefix_space);
    printf("--");
    for(j=0;j<A->size2-1;j++)
        printf("           ");
    printf("          --\n");
    
    free(prefix_space);
    return 0;
}

int print_vector(gsl_vector * p, char * prefix) {
    gsl_matrix_view A = gsl_matrix_view_vector(p, p->size, 1);
    return print_matrix(&A.matrix, prefix);
}

/* Read count double values into array data
 * Fills with zeros if too few values given */ 
int read_double_array(double * data, int count) {
    char input[20*count];
    char * endptr = NULL;
    char * startptr = &input[0];
    double number;
    int readCount = 0;
    errno = 0;
    
    if ( fgets(input, sizeof input, stdin) != NULL ) {
        while (1) {
            if (*startptr==',' || *startptr=='[')
                *startptr=' ';
            
            if (startptr==NULL || *startptr=='\0' || *startptr=='\n'|| *startptr==']')
                break;
            number = strtod(startptr, &endptr);
            
            if(errno!=0 || startptr==endptr)
                break;
            
            if(readCount >= count) {
                printf("You entered too many values, discarding the rest (beginning with %5.2f)", number);
                break;                
            }
            
            *(data+readCount) = number;
            
            readCount++;
            
            startptr = endptr;
        }
        
    }
    
    if (readCount<count) {
        printf("You entered too few values, setting the rest to 0.0\n");
        for(;readCount<count;readCount++){
            *(data+readCount) = 0.0;
        }
    }
    return 0;    
}