/*
 *   lwrmodels - Dynamic Robot Models of the KUKA Lightweight Robot IV
 *   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
 *
 *   This file is part of lwrmodels.
 *
 *   lwrmodels is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <model.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

int read_double_array(double * data, int count);
int print_matrix(double * A, int n, int m, char * prefix);

int main() {
    double q[NUMJ_MOD];
    double dq[NUMJ_MOD];
    double ddq[NUMJ_MOD];
    
    double tau[NUMJ_MOD];
    
    double B[NUMJ_MOD][NUMJ_MOD];
    double C[NUMJ_MOD][NUMJ_MOD];    
    double g[NUMJ_MOD];
    double tau_f[NUMJ_MOD];
    
    int i,j;
    
    printf("Using a dynamic robot model for %i joints.\n", NUMJ_MOD);
    printf("These are the joints ");
    
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%i", SIMJOINTS[i]+1);
        
        if (i==NUMJ_MOD-2)
            printf(" and ");
        else if (i==NUMJ_MOD-1)
            printf(" ");
        else
            printf(", ");
    }
    printf("of a robot with %i joints.\n\n", NUMJ_ROB);
    
    printf("Enter %i joint positions (space separated): ", NUMJ_MOD);
    read_double_array(q, NUMJ_MOD); 
    printf("Enter %i joint velocities (space separated): ", NUMJ_MOD);
    read_double_array(dq, NUMJ_MOD);
    printf("Enter %i joint accelerations (space separated): ", NUMJ_MOD); 
    read_double_array(ddq, NUMJ_MOD);
    
    calc_gravity_torque_vector(g, q);    
    calc_friction_torque_vector(tau_f, q);
    calc_inertia_matrix(B, q);
    calc_coriolis_matrix(C, q, dq);
    
        
    printf("\n\nInput set to:\n");
    printf("      q = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", q[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T");
        else
            printf(", ");
    }
    printf("\n     dq = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", dq[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T");
        else
            printf(", ");
        
    }
    printf("\n    tau = [");
    for(i=0;i<NUMJ_MOD;i++) { 
        printf("%.3f", tau[i]);
        if (i==NUMJ_MOD-1)
            printf("]^T\n");
        else
            printf(", ");
        
    }
    printf("\n\n\nOutput:\n");
    
    
    print_matrix((double *)&B, NUMJ_MOD, NUMJ_MOD, "     B(q) = ");
    print_matrix((double *)&C, NUMJ_MOD, NUMJ_MOD, "  C(q,dq) = ");
    print_matrix((double *)&g, NUMJ_MOD, 1, "     g(q) = ");
    print_matrix((double *)&tau_f, NUMJ_MOD, 1, "tau_f(dq) = ");

    for (i=0;i<NUMJ_MOD;i++) {
        tau[i] = 0.0;
        for (j=0;j<NUMJ_MOD;j++) {
            tau[i] += B[i][j] * ddq[j] + C[i][j] * dq[j];
        }
        tau[i] += g[i];
        tau[i] += tau_f[i];
    }
        
    printf("Computed actutation torques required:\n");
    print_matrix((double *)&tau, NUMJ_MOD, 1, "tau = ");
    
    return 0;
}

/* Print a matrix, print prefix in front of it */
int print_matrix(double * A, int n, int m, char * prefix) {
    int i=0, j=0, k=0, midline=0, mod2=0;
    
    char * prefix_space = (char *)malloc((strlen(prefix)+1)*sizeof(char));
    
    mod2 = n % 2;
    midline = (n - mod2)/2 ;
    
    k=0;
    while(*(prefix+k)!='\0') {
        *(prefix_space+k) = ' ';
        k++;       
    }
    *(prefix_space+k) = '\0';
    
    printf("%s--", prefix_space);
        
    for(j=0;j<m-1;j++)
        printf("           ");
    printf("          --\n");
    
    for(i=0;i<n;i++) {
        if(i==midline)
            printf("%s", prefix);
        else
            printf("%s", prefix_space);
    
        printf("| ");
        if(i==midline && mod2==0) {
            for(j=0;j<m;j++) {
                    printf("          ");
                if (j<m-1)
                    printf(" ");
                else
                    printf(" |\n");
            }
            printf("%s", prefix_space);
            printf("| ");
        }
        for(j=0;j<m;j++) {
            printf("%10.3e", *(A+i*m+j));
            
            if (j<m-1)
                printf(" ");
            else
                printf(" |\n");
        }
    }
    printf("%s", prefix_space);
    printf("--");
    for(j=0;j<m-1;j++)
        printf("           ");
    printf("          --\n");
    
    free(prefix_space);
    return 0;
}

/* Read count double values into array data
 * Fills with zeros if too few values given */
int read_double_array(double * data, int count) {
    char input[20*count];
    char * endptr = NULL;
    char * startptr = &input[0];
    double number;
    int readCount = 0;
    errno = 0;
    
    if ( fgets(input, sizeof input, stdin) != NULL ) {
        while (1) {
            if (*startptr==',' || *startptr=='[')
                *startptr=' ';
            
            if (startptr==NULL || *startptr=='\0' || *startptr=='\n'|| *startptr==']')
                break;
            number = strtod(startptr, &endptr);
            
            if(errno!=0 || startptr==endptr)
                break;
            
            if(readCount >= count) {
                printf("You entered too many values, discarding the rest (beginning with %5.2f)", number);
                break;                
            }
            
            *(data+readCount) = number;
            
            readCount++;
            
            startptr = endptr;
        }
        
    }
    
    if (readCount<count) {
        printf("You entered too few values, setting the rest to 0.0\n");
        for(;readCount<count;readCount++){
            *(data+readCount) = 0.0;
        }
    }
    return 0;    
}