/*
 *   lwrmodels - Dynamic Robot Models of the KUKA Lightweight Robot IV
 *   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
 *
 *   This file is part of lwrmodels.
 *
 *   lwrmodels is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include "model.h"

int SIMJOINTS[NUMJ_MOD] = {0, 1, 3};

void calc_residual(double resvec[2*NUMJ_MOD], const double q[NUMJ_MOD], 
                   const double dq[NUMJ_MOD], const double dqp[NUMJ_MOD], 
                   const double ddqp[NUMJ_MOD], const double tau[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD], dqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
        dqsq[i] = pow(dq[i],2);    
    }
 
    resvec[0] = dq[0] - dqp[0];
    resvec[1] = -0.0168799610481578*ddqp[0]*sin(2*q[1]) + 0.53524835732352*ddqp[0]*cosq[2] - 0.626240578068518*ddqp[0]*cos(2*q[1] - 2*q[2]) - 0.53524835732352*ddqp[0]*cos(2*q[1] - q[2]) - 1.57199673626875*ddqp[0] + 0.0175195488386143*ddqp[1]*cosq[1] + 1.25248115613704*dq[0]*dq[1]*sin(2*q[1] - 2*q[2]) + 1.07049671464704*dq[0]*dq[1]*sin(2*q[1] - q[2]) - 0.0337599220963156*dq[0]*dq[1]*cos(2*q[1]) - 0.53524835732352*dq[0]*dq[2]*sinq[2] - 1.25248115613704*dq[0]*dq[2]*sin(2*q[1] - 2*q[2]) - 0.53524835732352*dq[0]*dq[2]*sin(2*q[1] - q[2]) - 0.0175195488386143*dqsq[1]*sinq[1] - 0.0204978721411787*dqsq[1]*sin(q[1] - q[2]) + 0.0409957442823574*dq[1]*dq[2]*sin(q[1] - q[2]) - 0.0204978721411787*dqsq[2]*sin(q[1] - q[2]) + 0.000175195488386143*(117*ddqp[1] + (-117)*ddqp[2])*cos(q[1] - q[2]) + 2.6491674286195*(ddqp[0]*sinq[1] + 2*dq[0]*dq[1]*cosq[1])*sinq[1] + 1.43009916890323*(ddqp[0]*cosq[1] - 2*dq[0]*dq[1]*sinq[1])*cosq[1] + 0.791116620901176*(ddqp[0]*cos(q[1] - q[2]) + 2*dq[0]*(-dq[1] + dq[2])*sin(q[1] - q[2]))*cos(q[1] - q[2]) + 0.186461043875677*atan(100*dq[0]);
    resvec[2] = dq[1] - dqp[1];
    resvec[3] = 0.0175195488386143*ddqp[0]*cosq[1] + 0.0204978721411787*ddqp[0]*cos(q[1] - q[2]) + 1.81771647041588*ddqp[1] - 0.422954084243519*ddqp[2] - 1.21906825971627*dqsq[0]*sinq[1]*cosq[1] - 0.23068226761793*dqsq[0]*sin(2*q[1] - 2*q[2]) - 0.53524835732352*dqsq[0]*sin(2*q[1] - q[2]) + 0.0168799610481578*dqsq[0]*cos(2*q[1]) - 1.07049671464704*dq[1]*dq[2]*sinq[2] + 0.53524835732352*dqsq[2]*sinq[2] + 0.0026762417866176*(400*ddqp[1] + (-200)*ddqp[2])*cosq[2] - 32.4854155934467*sinq[1] - 13.1269659633593*sin(q[1] - q[2]) + 0.41398104470607*cosq[1] + 0.184489618216264*atan(100*dq[1]);
    resvec[4] = dq[2] - dqp[2];
    resvec[5] = -0.0204978721411787*ddqp[0]*cos(q[1] - q[2]) - 0.53524835732352*ddqp[1]*cosq[2] - 0.422954084243519*ddqp[1] + 0.422954084243519*ddqp[2] + 0.26762417866176*dqsq[0]*sinq[2] + 0.23068226761793*dqsq[0]*sin(2*q[1] - 2*q[2]) + 0.26762417866176*dqsq[0]*sin(2*q[1] - q[2]) + 0.53524835732352*dqsq[1]*sinq[2] + 13.1269659633593*sin(q[1] - q[2]) + 0.156066608371045*atan(100*dq[2]);

    for (i=0;i<NUMJ_MOD;i++) {
        resvec[i*2+1] = resvec[i*2+1] - tau[i];
    }
}

void calc_jacobian(double resmat[2*NUMJ_MOD][2*NUMJ_MOD], const double cj, 
                   const double q[NUMJ_MOD], const double dq[NUMJ_MOD], 
                   const double dqp[NUMJ_MOD], const double ddqp[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD], dqsq[NUMJ_MOD];
    int i;

    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
        dqsq[i] = pow(dq[i],2);    
    }
 
    resmat[0][0] = -cj;
    resmat[1][0] = 0;
    resmat[2][0] = 0;
    resmat[3][0] = 0;
    resmat[4][0] = 0;
    resmat[5][0] = 0;
    resmat[0][1] = 1;
    resmat[1][1] = cj*(2.6491674286195*sinqsq[1] - 0.0168799610481578*sin(2*q[1]) + 1.43009916890323*cosqsq[1] + 0.53524835732352*cosq[2] + 0.791116620901176*pow(cos(q[1] - q[2]), 2) - 0.626240578068518*cos(2*q[1] - 2*q[2]) - 0.53524835732352*cos(2*q[1] - q[2]) - 1.57199673626875) + 2.43813651943255*dq[1]*sinq[1]*cosq[1] + 1.25248115613704*dq[1]*sin(2*q[1] - 2*q[2]) + 1.07049671464704*dq[1]*sin(2*q[1] - q[2]) - 0.0337599220963156*dq[1]*cos(2*q[1]) - 0.53524835732352*dq[2]*sinq[2] - 1.25248115613704*dq[2]*sin(2*q[1] - 2*q[2]) - 0.53524835732352*dq[2]*sin(2*q[1] - q[2]) + 1.58223324180235*(-dq[1] + dq[2])*sin(q[1] - q[2])*cos(q[1] - q[2]) + 18.6461043875677/(10000*dqsq[0] + 1);
    resmat[2][1] = 0;
    resmat[3][1] = cj*(0.0175195488386143*cosq[1] + 0.0204978721411787*cos(q[1] - q[2])) - 2.43813651943255*dq[0]*sinq[1]*cosq[1] - 0.461364535235861*dq[0]*sin(2*q[1] - 2*q[2]) - 1.07049671464704*dq[0]*sin(2*q[1] - q[2]) + 0.0337599220963156*dq[0]*cos(2*q[1]);
    resmat[4][1] = 0;
    resmat[5][1] = -0.0204978721411787*cj*cos(q[1] - q[2]) + 0.53524835732352*dq[0]*sinq[2] + 0.461364535235861*dq[0]*sin(2*q[1] - 2*q[2]) + 0.53524835732352*dq[0]*sin(2*q[1] - q[2]);
    resmat[0][2] = 0;
    resmat[1][2] = 1.25248115613704*ddqp[0]*sin(2*q[1] - 2*q[2]) + 1.07049671464704*ddqp[0]*sin(2*q[1] - q[2]) - 0.0337599220963156*ddqp[0]*cos(2*q[1]) - 0.0175195488386143*ddqp[1]*sinq[1] + 0.0675198441926312*dq[0]*dq[1]*sin(2*q[1]) + 2.50496231227407*dq[0]*dq[1]*cos(2*q[1] - 2*q[2]) + 2.14099342929408*dq[0]*dq[1]*cos(2*q[1] - q[2]) - 2.50496231227407*dq[0]*dq[2]*cos(2*q[1] - 2*q[2]) - 1.07049671464704*dq[0]*dq[2]*cos(2*q[1] - q[2]) - 0.0175195488386143*dqsq[1]*cosq[1] - 0.0204978721411787*dqsq[1]*cos(q[1] - q[2]) + 0.0409957442823574*dq[1]*dq[2]*cos(q[1] - q[2]) - 0.0204978721411787*dqsq[2]*cos(q[1] - q[2]) - 0.000175195488386143*(117*ddqp[1] + (-117)*ddqp[2])*sin(q[1] - q[2]) + 1.43009916890323*(-ddqp[0]*sinq[1] - 2*dq[0]*dq[1]*cosq[1])*cosq[1] + 2.6491674286195*(ddqp[0]*sinq[1] + 2*dq[0]*dq[1]*cosq[1])*cosq[1] + 0.791116620901176*(-ddqp[0]*sin(q[1] - q[2]) + 2*dq[0]*(-dq[1] + dq[2])*cos(q[1] - q[2]))*cos(q[1] - q[2]) + 1.21906825971627*(ddqp[0]*cosq[1] - 2*dq[0]*dq[1]*sinq[1])*sinq[1] - 0.791116620901176*(ddqp[0]*cos(q[1] - q[2]) + 2*dq[0]*(-dq[1] + dq[2])*sin(q[1] - q[2]))*sin(q[1] - q[2]);
    resmat[2][2] = -cj;
    resmat[3][2] = -0.0175195488386143*ddqp[0]*sinq[1] - 0.0204978721411787*ddqp[0]*sin(q[1] - q[2]) + 1.21906825971627*dqsq[0]*sinqsq[1] - 0.0337599220963156*dqsq[0]*sin(2*q[1]) - 1.21906825971627*dqsq[0]*cosqsq[1] - 0.461364535235861*dqsq[0]*cos(2*q[1] - 2*q[2]) - 1.07049671464704*dqsq[0]*cos(2*q[1] - q[2]) - 0.41398104470607*sinq[1] - 32.4854155934467*cosq[1] - 13.1269659633593*cos(q[1] - q[2]);
    resmat[4][2] = 0;
    resmat[5][2] = 0.0204978721411787*ddqp[0]*sin(q[1] - q[2]) + 0.461364535235861*dqsq[0]*cos(2*q[1] - 2*q[2]) + 0.53524835732352*dqsq[0]*cos(2*q[1] - q[2]) + 13.1269659633593*cos(q[1] - q[2]);
    resmat[0][3] = 0;
    resmat[1][3] = cj*(0.0175195488386143*cosq[1] + 0.0204978721411787*cos(q[1] - q[2])) + 2.43813651943255*dq[0]*sinq[1]*cosq[1] - 1.58223324180235*dq[0]*sin(q[1] - q[2])*cos(q[1] - q[2]) + 1.25248115613704*dq[0]*sin(2*q[1] - 2*q[2]) + 1.07049671464704*dq[0]*sin(2*q[1] - q[2]) - 0.0337599220963156*dq[0]*cos(2*q[1]) - 0.0350390976772286*dq[1]*sinq[1] - 0.0409957442823574*dq[1]*sin(q[1] - q[2]) + 0.0409957442823574*dq[2]*sin(q[1] - q[2]);
    resmat[2][3] = 1;
    resmat[3][3] = cj*(1.07049671464704*cosq[2] + 1.81771647041588) - 1.07049671464704*dq[2]*sinq[2] + 18.4489618216264/(10000*dqsq[1] + 1);
    resmat[4][3] = 0;
    resmat[5][3] = cj*(-0.53524835732352*cosq[2] - 0.422954084243519) + 1.07049671464704*dq[1]*sinq[2];
    resmat[0][4] = 0;
    resmat[1][4] = -0.53524835732352*ddqp[0]*sinq[2] - 1.25248115613704*ddqp[0]*sin(2*q[1] - 2*q[2]) - 0.53524835732352*ddqp[0]*sin(2*q[1] - q[2]) - 2.50496231227407*dq[0]*dq[1]*cos(2*q[1] - 2*q[2]) - 1.07049671464704*dq[0]*dq[1]*cos(2*q[1] - q[2]) - 0.53524835732352*dq[0]*dq[2]*cosq[2] + 2.50496231227407*dq[0]*dq[2]*cos(2*q[1] - 2*q[2]) + 0.53524835732352*dq[0]*dq[2]*cos(2*q[1] - q[2]) + 0.0204978721411787*dqsq[1]*cos(q[1] - q[2]) - 0.0409957442823574*dq[1]*dq[2]*cos(q[1] - q[2]) + 0.0204978721411787*dqsq[2]*cos(q[1] - q[2]) + 0.000175195488386143*(117*ddqp[1] + (-117)*ddqp[2])*sin(q[1] - q[2]) + 0.791116620901176*(ddqp[0]*sin(q[1] - q[2]) - 2*dq[0]*(-dq[1] + dq[2])*cos(q[1] - q[2]))*cos(q[1] - q[2]) + 0.791116620901176*(ddqp[0]*cos(q[1] - q[2]) + 2*dq[0]*(-dq[1] + dq[2])*sin(q[1] - q[2]))*sin(q[1] - q[2]);
    resmat[2][4] = 0;
    resmat[3][4] = 0.0204978721411787*ddqp[0]*sin(q[1] - q[2]) + 0.461364535235861*dqsq[0]*cos(2*q[1] - 2*q[2]) + 0.53524835732352*dqsq[0]*cos(2*q[1] - q[2]) - 1.07049671464704*dq[1]*dq[2]*cosq[2] + 0.53524835732352*dqsq[2]*cosq[2] - 0.0026762417866176*(400*ddqp[1] + (-200)*ddqp[2])*sinq[2] + 13.1269659633593*cos(q[1] - q[2]);
    resmat[4][4] = -cj;
    resmat[5][4] = -0.0204978721411787*ddqp[0]*sin(q[1] - q[2]) + 0.53524835732352*ddqp[1]*sinq[2] + 0.26762417866176*dqsq[0]*cosq[2] - 0.461364535235861*dqsq[0]*cos(2*q[1] - 2*q[2]) - 0.26762417866176*dqsq[0]*cos(2*q[1] - q[2]) + 0.53524835732352*dqsq[1]*cosq[2] - 13.1269659633593*cos(q[1] - q[2]);
    resmat[0][5] = 0;
    resmat[1][5] = -0.0204978721411787*cj*cos(q[1] - q[2]) - 0.53524835732352*dq[0]*sinq[2] + 1.58223324180235*dq[0]*sin(q[1] - q[2])*cos(q[1] - q[2]) - 1.25248115613704*dq[0]*sin(2*q[1] - 2*q[2]) - 0.53524835732352*dq[0]*sin(2*q[1] - q[2]) + 0.0409957442823574*dq[1]*sin(q[1] - q[2]) - 0.0409957442823574*dq[2]*sin(q[1] - q[2]);
    resmat[2][5] = 0;
    resmat[3][5] = cj*(-0.53524835732352*cosq[2] - 0.422954084243519) - 1.07049671464704*dq[1]*sinq[2] + 1.07049671464704*dq[2]*sinq[2];
    resmat[4][5] = 1;
    resmat[5][5] = 0.422954084243519*cj + 15.6066608371045/(10000*dqsq[2] + 1);

}

void calc_inertia_matrix(double resmat[NUMJ_MOD][NUMJ_MOD], const double q[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
    }
 
   resmat[0][0] = 2.6491674286195*sinqsq[1] - 0.0168799610481578*sin(2*q[1]) + 1.43009916890323*cosqsq[1] + 0.53524835732352*cosq[2] + 0.791116620901176*pow(cos(q[1] - q[2]), 2) - 0.626240578068518*cos(2*q[1] - 2*q[2]) - 0.53524835732352*cos(2*q[1] - q[2]) - 1.57199673626875;
   resmat[1][0] = 0.0175195488386143*cosq[1] + 0.0204978721411787*cos(q[1] - q[2]);
   resmat[2][0] = -0.0204978721411787*cos(q[1] - q[2]);
   resmat[0][1] = 0.0175195488386143*cosq[1] + 0.0204978721411787*cos(q[1] - q[2]);
   resmat[1][1] = 1.07049671464704*cosq[2] + 1.81771647041588;
   resmat[2][1] = -0.53524835732352*cosq[2] - 0.422954084243519;
   resmat[0][2] = -0.0204978721411787*cos(q[1] - q[2]);
   resmat[1][2] = -0.53524835732352*cosq[2] - 0.422954084243519;
   resmat[2][2] = 0.422954084243519;

}

void calc_gravity_torque_vector(double resvec[NUMJ_MOD], const double q[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD];
    int i;

    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
    }
 
   resvec[0] = 0;
   resvec[1] = -32.4854155934467*sinq[1] - 13.1269659633593*sin(q[1] - q[2]) + 0.41398104470607*cosq[1];
   resvec[2] = 13.1269659633593*sin(q[1] - q[2]);

}

void calc_friction_torque_vector(double resvec[NUMJ_MOD], const double dq[NUMJ_MOD]) { 
   resvec[0] = 0.186461043875677*atan(100*dq[0]);
   resvec[1] = 0.184489618216264*atan(100*dq[1]);
   resvec[2] = 0.156066608371045*atan(100*dq[2]);

}

void calc_coriolis_matrix(double resmat[NUMJ_MOD][NUMJ_MOD], const double q[NUMJ_MOD], const double dq[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);  
    }
 
   resmat[0][0] = 0.609534129858137*dq[1]*sin(2*q[1]) + 0.23068226761793*dq[1]*sin(2*q[1] - 2*q[2]) + 0.53524835732352*dq[1]*sin(2*q[1] - q[2]) - 0.0168799610481578*dq[1]*cos(2*q[1]) - 0.26762417866176*dq[2]*sinq[2] - 0.23068226761793*dq[2]*sin(2*q[1] - 2*q[2]) - 0.26762417866176*dq[2]*sin(2*q[1] - q[2]);
   resmat[1][0] = -(-1)*dq[0]*(-152.383532464534*sin(2*q[1]) - 57.6705669044826*sin(2*q[1] - 2*q[2]) - 133.81208933088*sin(2*q[1] - q[2]) + 4.21999026203945*cos(2*q[1]))/250;
   resmat[2][0] = -(-1)*(-1)*dq[0]*(-66.90604466544*sinq[2] - 57.6705669044826*sin(2*q[1] - 2*q[2]) - 66.90604466544*sin(2*q[1] - q[2]))/250;
   resmat[0][1] = -(-1)*(-1)*dq[0]*(-152.383532464534*sin(2*q[1]) - 57.6705669044826*sin(2*q[1] - 2*q[2]) - 133.81208933088*sin(2*q[1] - q[2]) + 4.21999026203945*cos(2*q[1]))/250 + dq[1]*(-0.0175195488386143*sinq[1] - 0.0204978721411787*sin(q[1] - q[2])) + 0.0204978721411787*dq[2]*sin(q[1] - q[2]);
   resmat[1][1] = -0.53524835732352*dq[2]*sinq[2];
   resmat[2][1] = 0.53524835732352*dq[1]*sinq[2];
   resmat[0][2] = -0.26762417866176*dq[0]*sinq[2] - 0.23068226761793*dq[0]*sin(2*q[1] - 2*q[2]) - 0.26762417866176*dq[0]*sin(2*q[1] - q[2]) + 0.0204978721411787*dq[1]*sin(q[1] - q[2]) - 0.0204978721411787*dq[2]*sin(q[1] - q[2]);
   resmat[1][2] = 1.3381208933088*(-(-1)*(-2)*dq[1]/5 - (-1)*2*dq[2]/5)*sinq[2];
   resmat[2][2] = 0;

}
