/*
 *   lwrmodels - Dynamic Robot Models of the KUKA Lightweight Robot IV
 *   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
 *
 *   This file is part of lwrmodels.
 *
 *   lwrmodels is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include "model.h"

int SIMJOINTS[NUMJ_MOD] = {1, 3};

void calc_residual(double resvec[2*NUMJ_MOD], const double q[NUMJ_MOD], 
                   const double dq[NUMJ_MOD], const double dqp[NUMJ_MOD], 
                   const double ddqp[NUMJ_MOD], const double tau[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD], dqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
        dqsq[i] = pow(dq[i],2);    
    }
 
    resvec[0] = dq[0] - dqp[0];
    resvec[1] = 1.747789839512*ddqp[0] - 0.443227999912*ddqp[1] - 1.0645822136*dq[0]*dq[1]*sinq[1] + 0.5322911068*dqsq[1]*sinq[1] - 0.0068704576*dq[1]*(-2*dq[0] + dq[1])*cosq[1] - 0.0068704576*(-2*ddqp[0] + ddqp[1])*sinq[1] + 0.002661455534*(400*ddqp[0] + (-200)*ddqp[1])*cosq[1] - 32.52745327032*sinq[0] - 13.05443939427*sin(q[0] - q[1]) + 0.49870362231*cosq[0] - 0.16849797264*cos(q[0] - q[1]) + 0.145968455*atan(100*dq[0]);
    resvec[2] = dq[1] - dqp[1];
    resvec[3] = -0.0068704576*ddqp[0]*sinq[1] - 0.5322911068*ddqp[0]*cosq[1] - 0.443227999912*ddqp[0] + 0.443227999912*ddqp[1] + 0.5322911068*dqsq[0]*sinq[1] - 0.0068704576*dqsq[0]*cosq[1] + 13.05443939427*sin(q[0] - q[1]) + 0.16849797264*cos(q[0] - q[1]) + 0.156229968*atan(100*dq[1]);

    for (i=0;i<NUMJ_MOD;i++) {
        resvec[i*2+1] = resvec[i*2+1] - tau[i];
    }
}

void calc_jacobian(double resmat[2*NUMJ_MOD][2*NUMJ_MOD], const double cj, 
                   const double q[NUMJ_MOD], const double dq[NUMJ_MOD], 
                   const double dqp[NUMJ_MOD], const double ddqp[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD], dqsq[NUMJ_MOD];
    int i;

    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
        dqsq[i] = pow(dq[i],2);    
    }
 
    resmat[0][0] = -cj;
    resmat[1][0] = -0.49870362231*sinq[0] + 0.16849797264*sin(q[0] - q[1]) - 32.52745327032*cosq[0] - 13.05443939427*cos(q[0] - q[1]);
    resmat[2][0] = 0;
    resmat[3][0] = -0.16849797264*sin(q[0] - q[1]) + 13.05443939427*cos(q[0] - q[1]);
    resmat[0][1] = 1;
    resmat[1][1] = cj*(0.0137409152*sinq[1] + 1.0645822136*cosq[1] + 1.747789839512) - 1.0645822136*dq[1]*sinq[1] + 0.0137409152*dq[1]*cosq[1] + 14.5968455/(10000*dqsq[0] + 1);
    resmat[2][1] = 0;
    resmat[3][1] = cj*(-0.0068704576*sinq[1] - 0.5322911068*cosq[1] - 0.443227999912) + 1.0645822136*dq[0]*sinq[1] - 0.0137409152*dq[0]*cosq[1];
    resmat[0][2] = 0;
    resmat[1][2] = -1.0645822136*dq[0]*dq[1]*cosq[1] + 0.5322911068*dqsq[1]*cosq[1] + 0.0068704576*dq[1]*(-2*dq[0] + dq[1])*sinq[1] - 0.0068704576*(-2*ddqp[0] + ddqp[1])*cosq[1] - 0.002661455534*(400*ddqp[0] + (-200)*ddqp[1])*sinq[1] - 0.16849797264*sin(q[0] - q[1]) + 13.05443939427*cos(q[0] - q[1]);
    resmat[2][2] = -cj;
    resmat[3][2] = 0.5322911068*ddqp[0]*sinq[1] - 0.0068704576*ddqp[0]*cosq[1] + 0.0068704576*dqsq[0]*sinq[1] + 0.5322911068*dqsq[0]*cosq[1] + 0.16849797264*sin(q[0] - q[1]) - 13.05443939427*cos(q[0] - q[1]);
    resmat[0][3] = 0;
    resmat[1][3] = cj*(-0.0068704576*sinq[1] - 0.5322911068*cosq[1] - 0.443227999912) - 1.0645822136*dq[0]*sinq[1] + 1.0645822136*dq[1]*sinq[1] - 0.0068704576*dq[1]*cosq[1] - 0.0068704576*(-2*dq[0] + dq[1])*cosq[1];
    resmat[2][3] = 1;
    resmat[3][3] = 0.443227999912*cj + 15.6229968/(10000*dqsq[1] + 1);

}

void calc_inertia_matrix(double resmat[NUMJ_MOD][NUMJ_MOD], const double q[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);
    }
 
   resmat[0][0] = 0.0137409152*sinq[1] + 1.0645822136*cosq[1] + 1.747789839512;
   resmat[1][0] = -0.0068704576*sinq[1] - 0.5322911068*cosq[1] - 0.443227999912;
   resmat[0][1] = -0.0068704576*sinq[1] - 0.5322911068*cosq[1] - 0.443227999912;
   resmat[1][1] = 0.443227999912000;

}

void calc_gravity_torque_vector(double resvec[NUMJ_MOD], const double q[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD];
    int i;

    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
    }
 
   resvec[0] = -32.52745327032*sinq[0] - 13.05443939427*sin(q[0] - q[1]) + 0.49870362231*cosq[0] - 0.16849797264*cos(q[0] - q[1]);
   resvec[1] = 13.05443939427*sin(q[0] - q[1]) + 0.16849797264*cos(q[0] - q[1]);

}

void calc_friction_torque_vector(double resvec[NUMJ_MOD], const double dq[NUMJ_MOD]) { 
   resvec[0] = 0.145968455*atan(100*dq[0]);
   resvec[1] = 0.156229968*atan(100*dq[1]);

}

void calc_coriolis_matrix(double resmat[NUMJ_MOD][NUMJ_MOD], const double q[NUMJ_MOD], const double dq[NUMJ_MOD]) {
    double sinq[NUMJ_MOD], cosq[NUMJ_MOD], sinqsq[NUMJ_MOD], cosqsq[NUMJ_MOD];
    int i;
  
    for (i=0; i<NUMJ_MOD; i++) {
        cosq[i] = cos(q[i]);
        sinq[i] = sin(q[i]);
        cosqsq[i] = pow(cosq[i],2);
        sinqsq[i] = pow(sinq[i],2);  
    }
 
   resmat[0][0] = -(-1)*(-2)*dq[1]*(1.330727767*sinq[1] - 0.017176144*cosq[1])/5;
   resmat[1][0] = -(-1)*2*dq[0]*(1.330727767*sinq[1] - 0.017176144*cosq[1])/5;
   resmat[0][1] = (-(-1)*(-2)*dq[0]/5 - (-1)*2*dq[1]/5)*(1.330727767*sinq[1] - 0.017176144*cosq[1]);
   resmat[1][1] = 0;

}
