/*
 *   lwrsundialsmodels - Adapts system equations of lwrmodels to be usable
 *                       with SUNDIALS IDA solver for implicit ODEs
 *                      
 *   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
 *
 *   This file is part of lwrsundialsmodels.
 *
 *   lwrsundialsmodels is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include "sundialsmodel.h"
#include "model.h"
#include <ida/ida.h>
#include <ida/ida_dense.h>
#include <nvector/nvector_serial.h>
#include <sundials/sundials_math.h>
#include <sundials/sundials_types.h>

/* Problem Constants */

#define NOUT  1000

#define ZERO RCONST(0.0);
#define ONE  RCONST(1.0);



/* Prototypes of private functions */
static void PrintHeader(realtype rtol, N_Vector avtol, N_Vector yy, N_Vector yp);
static void PrintOutput(void *mem, realtype t, N_Vector y, double* tau);
static void PrintFinalStats(void *mem);
static int check_flag(void *flagvalue, char *funcname, int opt);

int read_double_array(double * data, int count);
/*
 *--------------------------------------------------------------------
 * Main Program
 *--------------------------------------------------------------------
 */


int main(void) {
    TauData dataval;
    TauData * data = &dataval;
    void *mem;
    N_Vector yy, yp, avtol, idv;
    realtype rtol, *idvval, *yval, *ypval, *atval, 
    q0[NUMJ_MOD], q[NUMJ_MOD], g[NUMJ_MOD], qref[NUMJ_MOD], qerr[NUMJ_MOD];
    realtype t0, tout, tret, Ts, tmp;
    int iout, retval, i;
    Ts = 0.01;
    //data = (UserData) malloc(sizeof *data);

    for(i=0;i<NUMJ_MOD;i++) {
        data->tau[i] = 0.0;
        qerr[i]=0.0;
    }

    mem = NULL;
    yy = yp = avtol = NULL;
    yval = ypval = atval = NULL;
    idvval = NULL;

    /* Allocate N-vectors. */
    yy = N_VNew_Serial(NUMJ_MOD*2);
    if(check_flag((void *)yy, "N_VNew_Serial", 0))
        return(1);
    
    yp = N_VNew_Serial(NUMJ_MOD*2);
    if(check_flag((void *)yp, "N_VNew_Serial", 0))
        return(1);
    
    avtol = N_VNew_Serial(NUMJ_MOD*2);
    if(check_flag((void *)avtol, "N_VNew_Serial", 0))
        return(1);
    
    idv = N_VNew_Serial(NUMJ_MOD*2);
    if(check_flag((void *)idv, "N_VNew_Serial", 0))
        return(1);

    /* Create and initialize  y, y', and absolute tolerance vectors. */
    yval  = NV_DATA_S(yy);
    ypval = NV_DATA_S(yp);
    idvval = NV_DATA_S(idv);

    for(i=0;i<NUMJ_MOD*2;i++) {
        yval[i] = 0.0;
        ypval[i] = 0.0;
        idvval[i] = 1.0;
    }

    if (NUMJ_MOD == 2)
        yval[2] = 90.0 / 180.0 * M_PI;
    else if (NUMJ_MOD == 7)
        yval[2] = 000.0 / 180.0 * M_PI;

    rtol = 1.0e-2;

    atval = NV_DATA_S(avtol);

    for(i=0;i<NUMJ_MOD;i++) {
        atval[i*2] = 1.0e-10;
        atval[i*2+1] = 1.0e-10;
        qref[i] = yval[i*2];
    }

    qref[1]=0.0;
    qref[3]=M_PI/2.0;
    /* Integration limits */
    t0 = ZERO;


    /* Call IDACreate and IDAMalloc to initialize IDA memory */
    mem = IDACreate();
    if(check_flag((void *)mem, "IDACreate", 0))
        return(1);
    
    retval = IDAInit(mem, resrob, t0, yy, yp);
    if(check_flag(&retval, "IDAInit", 1))
        return(1);
    
    retval = IDASVtolerances(mem, rtol, avtol);
    if(check_flag(&retval, "IDASVtolerances", 1)) 
        return(1);
    
    retval = IDASetUserData(mem, data);
    if(check_flag(&retval, "IDASetUserData", 1)) 
        return(1);
    
    retval = IDASetMaxNumSteps(mem, 50000);
    if(check_flag(&retval, "IDASetMaxNumSteps", 1)) 
        return(1);
    
    retval = IDASetMaxStep(mem, Ts);
    if(check_flag(&retval, "IDASetMaxStep", 1))
        return(1);

    /* Call IDADense and set up the linear solver. */
    retval = IDADense(mem, NUMJ_MOD*2);
    if(check_flag(&retval, "IDADense", 1))
        return(1);

    retval = IDADlsSetDenseJacFn(mem, jacrob);
    if(check_flag(&retval, "IDADlsSetDenseJacFn", 1))
        return(1);
            
    retval = IDASetId(mem, idv);
    if(check_flag(&retval, "IDASetId", 1))
        return(1);

    retval = IDACalcIC(mem, IDA_YA_YDP_INIT, t0 + 1.0);
    if(check_flag(&retval, "IDACalcIC", 1))
        return(1);

    retval = IDAGetConsistentIC(mem, yy, yp);
    if(check_flag(&retval, "IDAGetConsistentIC", 1))
        return(1);


    PrintHeader(rtol, avtol, yy, yp);
    /* Free avtol */
    N_VDestroy_Serial(avtol);
    /* In loop, call IDASolve, print results, and test for error.
        Break out of loop when NOUT preset output times have been reached. */

    iout = 0; tout = Ts;
    while(1) {

        //if(qref[1]<M_PI/20.0)
            //qref[1]+=0.5;
        
        for(i=0;i<NUMJ_MOD;i++) {
            q[i] = yval[i*2];
        }
        calc_gravity_torque_vector(g,q);
        //yval  = NV_DATA_S(yy);
        for(i=0;i<NUMJ_MOD;i++) {
            tmp = tout/1.0;
            if(tmp>1.0) tmp=1.0;
            tmp = qref[i]*tmp - yval[i*2];
            data->tau[i] = 1.25* tmp + 1.0*(tmp - qerr[i])/Ts +g[i];
            qerr[i] = tmp;
        }

        retval = IDASolve(mem, tout, &tret, yy, yp, IDA_NORMAL);

        PrintOutput(mem,tret,yy,data->tau);


        if(check_flag(&retval, "IDASolve", 1)) 
            return(1);

            if (retval == IDA_SUCCESS) {
                iout++;
                tout += Ts;
            }

            if (iout == NOUT) 
                break;
    }

    PrintFinalStats(mem);

    /* Free memory */

    IDAFree(&mem);
    N_VDestroy_Serial(yy);
    N_VDestroy_Serial(yp);

    return(0);

}

/*
 *--------------------------------------------------------------------
 * Private functions
 *--------------------------------------------------------------------
 */

/* 
 * Print first lines of output (problem description)
 */

static void PrintHeader(realtype rtol, N_Vector avtol, N_Vector yy, N_Vector yp)
{
   realtype *atval, *yval, *ypval;

    atval  = NV_DATA_S(avtol);
    yval  = NV_DATA_S(yy);
    ypval = NV_DATA_S(yp);
    long int y_len = NV_LENGTH_S(yy);
    int len = 0;
    long int i = 0;

    printf("################################################################\n#\n");
    printf("#    Simulation of robot motion dynamics using SUNDIALS IDA    #\n\n");
    printf("################################################################\n#\n");;
    #if defined(SUNDIALS_EXTENDED_PRECISION)
printf("CHECK PRECISION SETTING!\n");
    #elif defined(SUNDIALS_DOUBLE_PRECISION)
    printf("#    Tolerance parameters:\n#        rtol = %lg\n#        atol = [ ", rtol);
    for(i=0;i<NUMJ_MOD*2;i++) {
        if (i==NUMJ_MOD*2-1)
            printf("%lg ]^T\n#\n",atval[i]);
        else
            printf("%lg, ",atval[i]);
    }
    #else
printf("CHECK PRECISION SETTING!\n");
    #endif
    printf("#    Consistent initial conditions computed as:\n#         y0 = [ ");

    for(i=0;i<NUMJ_MOD*2;i++) {
        if (i==NUMJ_MOD*2-1)
            printf("%.4e ]^T\n#        yp0 = [ ",yval[i]);
        else
            printf("%.4e, ",yval[i]);
    }
    for(i=0;i<NUMJ_MOD*2;i++) {
        if (i==NUMJ_MOD*2-1)
            printf("%.4e ]^T\n#\n",ypval[i]);
        else
            printf("%.4e, ",ypval[i]);
    }
    len = 8 + 3 + 1 + 12 + 9 + 13*y_len;
    for(i=0;i<len;i++)
        printf("#");
    printf("\n");
    printf("    t     ");
    printf("n  ");
    printf("k ");
    printf("      h            ");

    for(i=0;i<y_len/2;i++)
        printf("     q%-2ld         dq%-2ld         tau%-2ld    ", i+1,i+1,i+1);

    printf("\n");
    for(i=0;i<len;i++)
        printf("#");
    printf("\n");
}

/*
 * Print Output
 */

static void PrintOutput(void *mem, realtype t, N_Vector y, double* tau)
{
    realtype *yval;
    int retval, kused;
    long int nst;
    realtype hused;
    long int y_len = NV_LENGTH_S(y);
    long int i = 0;
    yval  = NV_DATA_S(y);

    retval = IDAGetLastOrder(mem, &kused);
    check_flag(&retval, "IDAGetLastOrder", 1);
    retval = IDAGetNumSteps(mem, &nst);
    check_flag(&retval, "IDAGetNumSteps", 1);
    retval = IDAGetLastStep(mem, &hused);
    check_flag(&retval, "IDAGetLastStep", 1);
#if defined(SUNDIALS_EXTENDED_PRECISION)
    printf("CHECK PRECISION SETTING!\n");
#elif defined(SUNDIALS_DOUBLE_PRECISION)
    printf("%8.3f %3ld %1d %12.4e      ", t, nst, kused, hused);

    for(i=0;i<y_len/2;i++) {
        printf("%12.4e ", yval[i*2]);
        printf("%12.4e ", yval[i*2+1]);
        printf("%12.4e ", tau[i]);
    }
    printf("\n");
    
#else
    printf("CHECK PRECISION SETTING!\n");
#endif
}

/*
 * Print final integrator statistics
 */

static void PrintFinalStats(void *mem)
{
  int retval;
  long int nst, nni, nje, nre, nreLS, netf, ncfn, nge;

  retval = IDAGetNumSteps(mem, &nst);
  check_flag(&retval, "IDAGetNumSteps", 1);
  retval = IDAGetNumResEvals(mem, &nre);
  check_flag(&retval, "IDAGetNumResEvals", 1);
  retval = IDADlsGetNumJacEvals(mem, &nje);
  check_flag(&retval, "IDADlsGetNumJacEvals", 1);
  retval = IDAGetNumNonlinSolvIters(mem, &nni);
  check_flag(&retval, "IDAGetNumNonlinSolvIters", 1);
  retval = IDAGetNumErrTestFails(mem, &netf);
  check_flag(&retval, "IDAGetNumErrTestFails", 1);
  retval = IDAGetNumNonlinSolvConvFails(mem, &ncfn);
  check_flag(&retval, "IDAGetNumNonlinSolvConvFails", 1);
  retval = IDADlsGetNumResEvals(mem, &nreLS);
  check_flag(&retval, "IDADlsGetNumResEvals", 1);
  retval = IDAGetNumGEvals(mem, &nge);
  check_flag(&retval, "IDAGetNumGEvals", 1);

  printf("\nFinal Run Statistics: \n\n");
  printf("Number of steps                    = %ld\n", nst);
  printf("Number of residual evaluations     = %ld\n", nre+nreLS);
  printf("Number of Jacobian evaluations     = %ld\n", nje);
  printf("Number of nonlinear iterations     = %ld\n", nni);
  printf("Number of error test failures      = %ld\n", netf);
  printf("Number of nonlinear conv. failures = %ld\n", ncfn);
  printf("Number of root fn. evaluations     = %ld\n", nge);
}

/*
 * Check function return value...
 *   opt == 0 means SUNDIALS function allocates memory so check if
 *            returned NULL pointer
 *   opt == 1 means SUNDIALS function returns a flag so check if
 *            flag >= 0
 *   opt == 2 means function allocates memory so check if returned
 *            NULL pointer 
 */

static int check_flag(void *flagvalue, char *funcname, int opt)
{
  int *errflag;
  /* Check if SUNDIALS function returned NULL pointer - no memory allocated */
  if (opt == 0 && flagvalue == NULL) {
    fprintf(stderr, 
            "\nSUNDIALS_ERROR: %s() failed - returned NULL pointer\n\n", 
            funcname);
    return(1);
  } else if (opt == 1) {
    /* Check if flag < 0 */
    errflag = (int *) flagvalue;
    if (*errflag < 0) {
      fprintf(stderr, 
              "\nSUNDIALS_ERROR: %s() failed with flag = %d\n\n", 
              funcname, *errflag);
      return(1); 
    }
  } else if (opt == 2 && flagvalue == NULL) {
    /* Check if function returned NULL pointer - no memory allocated */
    fprintf(stderr, 
            "\nMEMORY_ERROR: %s() failed - returned NULL pointer\n\n", 
            funcname);
    return(1);
  }

  return(0);
}
/* Read count double values into array data
 * Fills with zeros if too few values given */
int read_double_array(double * data, int count) {
    char input[20*count];
    char * endptr = NULL;
    char * startptr = &input[0];
    double number;
    int readCount = 0;
    errno = 0;

    if ( fgets(input, sizeof input, stdin) != NULL ) {
        while (1) {
            if (*startptr==',' || *startptr=='[')
                *startptr=' ';

            if (startptr==NULL || *startptr=='\0' || *startptr=='\n'|| *startptr==']')
                break;
            number = strtod(startptr, &endptr);

            if(errno!=0 || startptr==endptr)
                break;

            if(readCount >= count) {
                printf("You entered too many values, discarding the rest (beginning with %5.2f)", number);
                break;
            }

            *(data+readCount) = number;

            readCount++;

            startptr = endptr;
        }

    }

    if (readCount<count) {
        printf("You entered too few values, setting the rest to 0.0\n");
        for(;readCount<count;readCount++){
            *(data+readCount) = 0.0;
        }
    }
    return 0;
}