#
#   python-lwrmodels - Python interface to the dynamic robot models of the
#                      lwrmodels C-library
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrmodels.
#
#   python-lwrmodels is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#cython: embedsignature=True
import numpy as np
# "cimport" is used to import special compile-time information
# about the numpy module (this is stored in a file numpy.pxd which is
# currently part of the Cython distribution).
cimport numpy as np
# We now need to fix a datatype for our arrays. I've used the variable
# DTYPE for this, which is assigned to the usual NumPy runtime
# type info object.
DTYPE = np.float
# "ctypedef" assigns a corresponding compile-time type to DTYPE_t. For
# every type in the numpy module there's a corresponding compile-time
# type with a _t-suffix.
ctypedef np.float_t DTYPE_t

numJointsModel = NUMJ_MOD
numJointsRobot = NUMJ_ROB
simulatedJoints = [0]*NUMJ_MOD

def get_gravity_torque_vector(
        np.ndarray[DTYPE_t, ndim=1, negative_indices=False] q):
    """Compute the vector of torques due to gravity

    Args:
        q (numpy.array): Vector of joint positions 
    
    Returns:
        Vector of torques (1-dimensional numpy.array)
    """

    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] resvec = \
        np.zeros((NUMJ_MOD,), dtype=DTYPE)
        
    calc_gravity_torque_vector(<double*>resvec.data, <double*>q.data)
    return resvec

def get_inertia_matrix(np.ndarray[DTYPE_t, ndim=1, negative_indices=False] q): 
    """Compute the inertia matrix for the given position

    Args:
        q (numpy.array): Vector of joint positions 
    
    Returns:
        Inertia Matrix (2-dimensional numpy.array)
    """

    cdef np.ndarray[DTYPE_t, ndim=2, negative_indices=False] resmat = \
        np.zeros((NUMJ_MOD, NUMJ_MOD), dtype=DTYPE)
    cdef int i, j
    cdef double cresmat[NUMJ_MOD][NUMJ_MOD]
    
    calc_inertia_matrix(cresmat, <double*>q.data)
    
    for i in range(NUMJ_MOD):
        for j in range(NUMJ_MOD):
            resmat[i,j] = cresmat[i][j]
       
    return resmat

def get_friction_torque_vector(
        np.ndarray[DTYPE_t, ndim=1, negative_indices=False] dq):
    """Compute the vector of torques due to friction

    Args:
        dq (numpy.array): Vector of joint velocities 
    
    Returns:
        Vector of torques (1-dimensional numpy.array)
    """

    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] resvec = \
        np.zeros((NUMJ_MOD,), dtype=DTYPE)

    calc_friction_torque_vector(<double*>resvec.data, <double*>dq.data)
    
    return resvec

def get_coriolis_matrix(np.ndarray[DTYPE_t, ndim=1, negative_indices=False] q,
                        np.ndarray[DTYPE_t, ndim=1, negative_indices=False] dq):
    """Compute the Coriolis-/Centrifugal Matrix

    Args:
        q (numpy.array): Vector of joint positions 
    
        dq (numpy.array): Vector of joint velocities 
    
    Returns:
        Coriolis-/Centrifugal Matrix (2-dimensional numpy.array)
    """
    
    cdef np.ndarray[DTYPE_t, ndim=2, negative_indices=False] resmat = \
        np.zeros((NUMJ_MOD, NUMJ_MOD), dtype=DTYPE)
    cdef int i, j
    cdef double cresmat[NUMJ_MOD][NUMJ_MOD]
    
    calc_coriolis_matrix(cresmat, <double*>q.data, <double*>dq.data)
    
    for i in range(NUMJ_MOD):
        for j in range(NUMJ_MOD):
            resmat[i,j] = cresmat[i][j]
       
    return resmat

def get_torque_vector(np.ndarray[DTYPE_t, ndim=1, negative_indices=False] q,
                      np.ndarray[DTYPE_t, ndim=1, negative_indices=False] dq,
                      np.ndarray[DTYPE_t, ndim=1, negative_indices=False] ddq):
    
    """Compute the vector of actuation torques required for the given motion

    Args:
        q (numpy.array): Vector of joint positions 

        dq (numpy.array): Vector of joint velocities

        ddq (numpy.array): Vector of joint acceleration 
    
    Returns:
        Vector of required actuation torques (1-dimensional numpy.array)
    """

    cdef np.ndarray[DTYPE_t, ndim=1, negative_indices=False] resvec = \
        np.zeros((NUMJ_MOD,), dtype=DTYPE)
    
    resvec = (np.dot(get_inertia_matrix(q), ddq) 
             + np.dot(get_coriolis_matrix(q, dq), dq)
             + get_gravity_torque_vector(q)
             + get_friction_torque_vector(dq))
            
    return resvec

 


