#
#   python-lwrsim - Simulates the robot motion dynamics using
#                   the dynamic robot models of the lwrmodels and
#                   lwrsundialsmodel library with SUNDIALS IDA.
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrsim.
#
#   python-lwrsim is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from lwrsim.lwr4_q2q4 import DynamicsSimulator

# Get number of joints of the robot
nRob = DynamicsSimulator.numJointsRobot
# Set the sampling time
Ts = 0.01
# Set inital joint positions
q0 = [0.0, 3.0, 0.0, -1.5, 0.0, 0.0, 0.0]
# Set inital joint velocities
dq0 = [0.0]*nRob
# Set absolute tolerance for the solver
abstol = 1.0e-6
# Set relative tolerance for the solver
reltol = 1.0e-5

# Now we can initialize the simulator / the ODE solver
ds = DynamicsSimulator(sampleTime=Ts, q0=q0, dq0=dq0,
                        abstol=abstol, reltol=reltol)

# define a list to save the time to
t = [0.0]
# define a list to save the lists of joint positions to
q = [q0]
# define a list to save the lists of joint velocities to
dq = [dq0]

# Initialize list of actuation torques (all zero)
tau = [0.0]*nRob

# Let us simulate for 1000 samples, i.e. 1000*Ts = 10s
for k in range(1000):
    # Simulate the robot motion to time T = T + Ts = k*Ts
    # with the supplied actuation torques tau
    ds.calc_next_state(tau)

    # Save the results
    t += [(k+1)*Ts]
    q += [ds.get_joint_positions()]
    dq += [ds.get_joint_velocities()]

try:
    import numpy as np
    import pylab as P

    # First let us convert the data to numpy arrays
    t = np.array(t)
    q = np.array(q)
    dq = np.array(dq)

    # Then plot the data (refer to matplotlib documentation)
    P.subplot(2,1,1)
    P.plot(t, q[:,1], label='q2')
    P.plot(t, q[:,3], label='q4')
    P.legend()

    P.subplot(2,1,2)
    P.plot(t, dq[:,1], label='dq2')
    P.plot(t, dq[:,3], label='dq4')
    P.legend()

    P.show()
except (ImportError):
    pass


# Assemble format string for values and the title line
fmtstr = ', '.join(["{:7.3f}" for n in range(len(q0))])
titlestr = ", ".join(["   t   "] 
                   + ["   q{}  ".format(n+1) for n in range(len(q0))]
                   + ["  dq{}  ".format(n+1) for n in range(len(q0))])

# Write the data to a file in Comma-Separated-Value format
with open("/tmp/lwrsim_plotdata.csv", 'wt') as f:
    f.write(titlestr + "\n")
    for k in range(len(t)):
        f.write("{:7.3f}, ".format(t[k]))
        f.write(fmtstr.format(*q[k]))
        f.write(", ")
        f.write(fmtstr.format(*dq[k]))
        f.write("\n")
        