#
#   python-lwrsim - Simulates the robot motion dynamics using
#                   the dynamic robot models of the lwrmodels and
#                   lwrsundialsmodel library with SUNDIALS IDA.
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrsim.
#
#   python-lwrsim is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from lwrsim.lwr4_q2q4 import DynamicsSimulator
from lwrmodels.lwr4_q2q4.dynamics import get_gravity_torque_vector as tau_g
from sundials import IDAError
import numpy as np
from numpy import pi

# Get number of joints of the robot
nRob = DynamicsSimulator.numJointsRobot
# Set the sampling time
Ts = 0.01
# Set inital joint positions
q0 = [0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0]
# Set inital joint velocities
dq0 = [0.0]*nRob
# Set absolute tolerance for the solver
abstol = 1.0e-6
# Set relative tolerance for the solver
reltol = 1.0e-5

# Now we can initialize the simulator / the ODE solver
ds = DynamicsSimulator(sampleTime=Ts, q0=q0, dq0=dq0,
                        abstol=abstol, reltol=reltol)

# Determine how many samples to simulated for 10s simulation
nsamples = int(4.0 / Ts) 

# Set the controller gains
# Using same setting for all joints here
Kp = np.ones((nRob,)) * 10.0
Ki = Kp * 0.0
Kd = 2.0 * np.sqrt(Kp) * 1.0

# Define the contant reference position
qref = np.array([0.0, 0.0, 0.0, 1.0*pi/2.0, 0.0, 0.0, 0.0])

# initialize an array to save the time to
t = np.zeros((nsamples, ))
# initialize an array to save the joint positions to
q = np.zeros((nsamples, nRob))
q[0,:] = q0[:]
# initialize an array to save the joint velocities to
dq = np.zeros((nsamples, nRob))
dq[0,:] = dq0[:]

# initialize an array to save the actuation torques to
tau = np.zeros((nsamples, nRob))

# initialize arrays for 
# the last position error
lasterr = qref - q[0,:]
# the current position error
error = np.zeros((nRob,))
# the integral value of the error
errorint = np.zeros((nRob,))
# the derivative value of the error
derror = np.zeros((nRob,))

# Let us simulate for nsamples samples, i.e. 1000*Ts = 10s for Ts=0.01s
for k in range(1,nsamples):
    # Compute errors etc.
    error[:] = qref - q[k-1,:]
    errorint += error
    derror[:] = (error-lasterr)/Ts
    lasterr[:] = error[:]
    
    # Compute the torque from feedback control
    cur_tau = (Kp * error 
             + Kd * derror 
             + Ki * errorint)
   
    # Add gravity compensation torque
    cur_tau[ds.simulatedJoints] += tau_g(q[k-1,ds.simulatedJoints])
    
    try:
        # Simulate the robot motion to time T = T + Ts = k*Ts
        # with the supplied actuation torques tau
        # Note that we convert the numpy array to a list
        ds.calc_next_state(list(cur_tau))
    except (IDAError):
        print("Simulation failed, see error message above.")
        break
        
    # Save the results
    t[k] = k*Ts
    q[k,:] = ds.get_joint_positions()[:]
    dq[k,:] = ds.get_joint_velocities()[:]
    tau[k-1,:] = cur_tau[:]

tau[-1,:] = tau[-2,:]

try:
    import pylab as P

    # Plot the data (refer to matplotlib documentation)
    P.subplot(3,1,1)
    P.plot(t, q[:,1], 'b', label='q2')
    P.plot(t[[0,-1]], [qref[1]]*2, 'b--', label='qref2')
    P.plot(t, q[:,3], 'g', label='q4')
    P.plot(t[[0,-1]], [qref[3]]*2, 'g--', label='qref4')
    P.legend()

    P.subplot(3,1,2)
    P.plot(t, dq[:,1], label='dq2')
    P.plot(t, dq[:,3], label='dq4')
    P.legend()

    P.subplot(3,1,3)
    P.plot(t, tau[:,1], label='tau2')
    P.plot(t, tau[:,3], label='tau4')
    P.legend()

    P.show()
    #P.savefig('lwrsim.png', bbox_inches='tight')
except (ImportError):
    pass

    
# Assemble format string for values and the title line
fmtstr = ', '.join(["{:7.3f}" for n in range(nRob)])
titlestr = ", ".join(["   t   "] 
                   + ["   q{}  ".format(n+1) for n in range(nRob)]
                   + ["  dq{}  ".format(n+1) for n in range(nRob)]           
                   + [" tau{}  ".format(n+1) for n in range(nRob)])

# Write the data to a file in Comma-Separated-Value format
with open("/tmp/lwrsim_plotdata.csv", 'wt') as f:
    f.write(titlestr + "\n")
    for k in range(len(t)):
        f.write("{:7.3f}, ".format(t[k]))
        f.write(fmtstr.format(*q[k,:]))
        f.write(", ")
        f.write(fmtstr.format(*dq[k,:]))
        f.write(", ")
        f.write(fmtstr.format(*tau[k,:]))
        f.write("\n")
