#
#   python-lwrsim - Simulates the robot motion dynamics using
#                   the dynamic robot models of the lwrmodels and
#                   lwrsundialsmodel library with SUNDIALS IDA.
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrsim.
#
#   python-lwrsim is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import time

import numpy as np
from numpy import pi

from fripy.server import FRIServer
from fripy.fricomm import (FRI_STATE_MON, FRI_STATE_CMD, FRI_CTRL_POSITION,
    FRI_CTRL_JNT_IMP, LBR_MNJ) 

                     
# Set the sampling time
Ts = 0.01
# Determine how many samples to simulate for 4s simulation
nsamples = int(4.0 / Ts) 

# Set the controller gains
# Using same setting for all joints here
Kp = np.ones((LBR_MNJ,)) * 10.0
Ki = Kp * 0.0
Kd = 2.0 * np.sqrt(Kp) * 1.0

# Define the contant reference position
qref = np.array([0.0, 0.0, 0.0, 1.0*pi/2.0, 0.0, 0.0, 0.0])

# initialize an array to save the time to
t = np.zeros((nsamples, ))
# initialize an array to save the joint positions to
q = np.zeros((nsamples, LBR_MNJ))
# initialize an array to save the actuation torques to
tau = np.zeros((nsamples, LBR_MNJ))

# initialize arrays for 
# the last position error
lasterr = np.zeros((LBR_MNJ,))
# the current position error
error = np.zeros((LBR_MNJ,))
# the integral value of the error
errorint = np.zeros((LBR_MNJ,))
# the derivative value of the error
derror = np.zeros((LBR_MNJ,))

# Initialize the FRIServer which connects 
# to the robot or the emulation (FRIClient)
# supply a receive timeout
fri = FRIServer(simulation=True,timeout=1.0)

k = 0

# Define a timeout to quit the program 
# if no communication is going on for some time
last_comm_time = time.time()
timeout = 10.0

while True:    
    # Try to receive data
    if fri.recv_data():
        last_comm_time = time.time()
    else:
        # Quit if communication timed out
        if time.time() > last_comm_time + timeout:        
            print("Quit due to communication timeout.")
            break
        # Otherwise try to receive again
        else:
            continue
    
    # Control the joint positions if we ware in command mode
    # limit to number of samples defined above
    if (fri.msr.intf.state is FRI_STATE_CMD) and (k < nsamples):
        # Save current data for plotting later on
        q[k,:] = fri.msr.data.msrJntPos[:]
        tau[k,:] = fri.msr.data.msrJntTrq[:]
        t[k] = k*Ts
    
        # Check that we use the Joint Impedance Control Strategy
        if (fri.msr.robot.control is FRI_CTRL_JNT_IMP):
            
            # Compute errors etc.
            error[:] = qref - q[k,:]
            errorint += error
            
            # Set last error to current error initially, 
            # so we avoid a large initial control torque
            # resulting from the derivative term
            if k == 0:
                lasterr[:] = error[:]

            derror[:] = (error-lasterr)/Ts
            lasterr[:] = error[:]
            
            # Compute the torque from feedback control
            # Note: Gravity is already compensated by the robot side
            cur_tau = (Kp * error 
                     + Kd * derror 
                     + Ki * errorint)
            
            # Hand torque over to FRIServer
            fri.ctl_jnt_imp(newJntTrq=cur_tau[:], sendData=False)
        
        k += 1
    
    # Quit if communication quality decreased
    if (fri.msr.intf.state is FRI_STATE_MON) and (k > 0):
        print("Closing connection because state decreased to monitor.")
        fri.cmd.krl.intData[1] = 2
        fri.send_data()
        break
        
    # Quit if we have simulated for all samples
    if k >= nsamples:
        print("Closing connection because number of samples reached.")
        fri.cmd.krl.intData[1] = 2
        fri.send_data()
        break
            
    # Quit if Robot side / KRL program wants to stop
    if fri.msr.krl.intData[0] == 2:
        print("Simulator requested to close connection.")
        fri.cmd.krl.intData[1] == 2
        fri.send_data()
        break
    
    # Send data to robot side
    fri.send_data()
    
try:
    import pylab as P

    # Plot the data (refer to matplotlib documentation)
    P.subplot(2,1,1)
    P.plot(t, q[:,1], 'b', label='q2')
    P.plot(t[[0,-1]], [qref[1]]*2, 'b--', label='qref2')
    P.plot(t, q[:,3], 'g', label='q4')
    P.plot(t[[0,-1]], [qref[3]]*2, 'g--', label='qref4')
    P.legend()

    P.subplot(2,1,2)
    P.plot(t, tau[:,1], label='tau2')
    P.plot(t, tau[:,3], label='tau4')
    P.legend()

    P.show()
    #P.savefig('lwrsim.png', bbox_inches='tight')
except (ImportError):
    pass

# Assemble format string for values and the title line
fmtstr = ', '.join(["{:7.3f}" for n in range(LBR_MNJ)])
titlestr = ", ".join(["   t   "] 
                + ["   q{}  ".format(n+1) for n in range(LBR_MNJ)]           
                + [" tau{}  ".format(n+1) for n in range(LBR_MNJ)])

# Write the data to a file in Comma-Separated-Value format
with open("/tmp/lwrsim_plotdata.csv", 'wt') as f:
    f.write(titlestr + "\n")
    for k in range(len(t)):
        f.write("{:7.3f}, ".format(t[k]))
        f.write(fmtstr.format(*q[k,:]))
        f.write(", ")
        f.write(fmtstr.format(*tau[k,:]))
        f.write("\n")
