#
#   python-lwrsim - Simulates the robot motion dynamics using
#                   the dynamic robot models of the lwrmodels and
#                   lwrsundialsmodel library with SUNDIALS IDA.
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrsim.
#
#   python-lwrsim is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import time
import numpy as np

from lwrsim.lwr4_q2q4 import DynamicsSimulator
from lwrmodels.lwr4_q2q4.dynamics import get_gravity_torque_vector as tau_g
from sundials import IDAError

from fripy.fricomm import FRI_CTRL_POSITION, FRI_CTRL_JNT_IMP, LBR_MNJ
from fripy.fricomm import FRI_STATE_CMD, FRI_QUALITY_PERFECT 
from fripy.client import FRIClient

# Set the sampling time
Ts = 0.01
# Set inital joint positions
q0 = [0.0, 0.5, 0.0, 0.0, 0.0, 0.0, 0.0]
# Set inital joint velocities
dq0 = [0.0]*LBR_MNJ
# Set absolute tolerance for the solver
abstol = 1.0e-6
# Set relative tolerance for the solver
reltol = 1.0e-5

# Now we can initialize the simulator / the ODE solver
ds = DynamicsSimulator(sampleTime=Ts, q0=q0, dq0=dq0,
                        abstol=abstol, reltol=reltol)

# Initialize the FRIClient which emulates the robot's network communication
fri = FRIClient(debug=False, verbose=True)

# As we want to allow the control on torque level 
# we use the joint impedance control strategy 
fri.set_control_strategy(FRI_CTRL_JNT_IMP)
# Open the interface. The communication works as fast as possible, 
# so the answer time argument is not important. However, if the
# simulation takes longer than this time, a timeout occurs.
# So we use sth. larger to allow a simulation slower than real-time
fri.open(ds.sampleTime*100.0)

# Also set the inital position of the interface data
fri.msr.data.msrJntPos = ds.get_joint_positions()[:]

# Quit after a after timeout seconds with no communication
last_comm_time = time.time()
timeout = 10.0

# Simulate with torques supplied by FRI and return joint positions via FRI
while True:   
    # Call fri.start (allow control) and 
    # go to command mode if quality is perfect
    if (fri.msr.intf.quality is  FRI_QUALITY_PERFECT and 
        fri.msr.intf.state is not FRI_STATE_CMD):
        fri.start()
    
    # Check that we use Joint Impedance Control strategy
    if  fri.msr.robot.control is FRI_CTRL_JNT_IMP:
        fri.msr.data.msrJntTrq = [0.0]*LBR_MNJ
        
        # Compute and compensate gravity torque as the real robot does
        gravpos = np.array(ds.get_joint_positions())[ds.simulatedJoints]
        gravtorque = tau_g(gravpos)
        
        for jnt in range(len(ds.simulatedJoints)):
            simjnt = ds.simulatedJoints[jnt]
            fri.msr.data.msrJntTrq[simjnt] += gravtorque[jnt]

        # If we are in command mode, apply the supplied torque addJntTrq
        # and simulate the motion dynamics
        # then update the joint positions in the fri.msr.data structure
        if fri.msr.intf.state is FRI_STATE_CMD:
            for jnt in range(LBR_MNJ):
                fri.msr.data.msrJntTrq[jnt] += fri.cmd.cmd.addJntTrq[jnt]

            try:
                ds.calc_next_state(fri.msr.data.msrJntTrq)
            except (IDAError):
                # Handle the solver exception
                # In most cases the supplied torques were inappropriate
                print("Warning: Simulation failed, see error message above.")
                # For now just tolerate this and pass
                # Here, you could also reset the simulation state 
                # to the previous position, brake the robot
                # and switch to monitor mode
                # or quit directly                
                
            fri.msr.data.msrJntPos = ds.get_joint_positions()
            
            last_comm_time = time.time()
                    
    elif fri.msr.robot.control is FRI_CTRL_POSITION:
        # If position control strategy is used handle it here
        # Should not be the case as we want to control
        # the robot on torque level
        if fri.msr.intf.state is FRI_STATE_CMD:
            fri.msr.data.msrJntPos = fri.cmd.cmd.jntPos[:]
            ds.set_state(fri.msr.data.msrJntPos, [0.0]*LBR_MNJ, False)
           
    # Here we handle a timeout to quit our application if communication
    # failed for some time
    if time.time() > last_comm_time + timeout:
        print("Simulator closing due to timeout.")
        fri.msr.krl.intData[0] = 2
        fri.send_data()
        fri.close()
        break
    
    # Quit if remote side wants to stop
    if fri.cmd.krl.intData[1] == 2:        
        print("Stop by server requested")
        fri.msr.krl.intData[0] = 2
        fri.send_data()
        fri.close()
        break
    
    # Send and receive data via FRI
    fri.send_data()    
    fri.recv_data()
    
    # This handles a bug
    # If we receive an old package just receive again to catch up
    if fri._dont_answer:
        fri._dont_answer = False
        fri.recv_data()
