#
#   python-lwrsim - Simulates the robot motion dynamics using
#                   the dynamic robot models of the lwrmodels and
#                   lwrsundialsmodel library with SUNDIALS IDA.
#   Copyright (C) 2012 Vinzenz Bargsten <vb@bargsten.com>
#
#   This file is part of python-lwrsim.
#
#   python-lwrsim is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#   This file incorporates work covered by the following copyright and  
#   permission notice:  
#
#    Copyright (c) 2010 Runar Tenfjord. All rights reserved.
#
#    Redistribution and use in source and binary forms, with or without
#    modification, are permitted provided that the following conditions are
#    met:
#
#    1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#    
#    2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in
#        the documentation and/or other materials provided with the
#        distribution.
#
#    3. Neither the name of Tenko nor the names of its contributors may
#        be used to endorse or promote products derived from this software
#        without specific prior written permission.
#
#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INFRAE OR
#    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
#    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
#    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
#    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
#    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#cython: embedsignature=True
"""
.. module:: dynsim
   :synopsis: Simulates robot motion using SUNDIALS IDA.

.. moduleauthor:: Vinzenz Bargsten <vb@bargsten.com>

"""


from sundials.sundials cimport N_Vector_Serial, SolverStatistics,\
    N_Vector_data, N_Vector_length, IDASettings
from sundials.SundialsLib cimport *
from sundials.sundials import N_Vector_Serial_Copy_Array, IDAError
 
cdef class DynamicsSimulator(object):
    """
    This class interconntects the ODE solver SUNDIALS IDA, one of the
    dynamic robot models via the lwrsundialsmodels library and Python. It
    does most of the setup of the ODE solver and provides robot-specific
    functions to simulate and to get the results.

    Args:
        sampleTime (float): The sampling time to use. A single simulation run
                            will always simulate for this long, in [s].

        q0 (list of floats): List of initial joint positions, in [rad].

        dq0 (list of floats): List of initial joint velocities, in [rad/s].

        makeConsistent (bool): Compute consistent inital values for the solver.
                               In general recommended. See also SUNDIALS IDA
                               documentation.

    Kwargs:
        abstol (float): Absolute tolerance setting for the solver (required)

        reltol (float): Relative tolerance setting for the solver (required)

        mxsteps (int): Maximum number of internal steps allowed to be
                       calculated within a single simulation run 

        .. note:: Even if the model only describes some joints of the
            robot, the simulator expects and returns lists with
            values for all joints of the robot.
    """

    numJointsModel = NUMJ_MOD
    numJointsRobot = NUMJ_ROB
    simulatedJoints = [0]*NUMJ_MOD

    for i in range(NUMJ_MOD):
        simulatedJoints[i] = SIMJOINTS[i]

    def __init__(self, sampleTime=0.01, q0=None, dq0=None, 
        makeConsistent=True, **kwargs):

        self.y = N_Vector_Serial(2*NUMJ_MOD)
        self.ydot = N_Vector_Serial(2*NUMJ_MOD)
        
        for i in range(2*NUMJ_MOD):
            self.ydot[i] = 0.0
        
        if q0 is not None:
            for i in range(NUMJ_MOD):
                self.y[i*2] = float(q0[SIMJOINTS[i]])

        if dq0 is not None:
            for i in range(NUMJ_MOD):
                self.y[i*2+1] = float(dq0[SIMJOINTS[i]])
        
        print("Supplied Initial Conditions:")
        for i in range(2*NUMJ_MOD):
            print("y[{}]={}, ydot[{}]={}".format(i, self.y[i], i, self.ydot[i]))

        self.settings = IDASettings(**kwargs)
        self.sampleTime = sampleTime
        self.settings.hmax = float(sampleTime)
        

        
        print("Simulating Joints:")
        print(DynamicsSimulator.simulatedJoints)
        
        self._init_solver(makeConsistent=makeConsistent)

    def __dealloc__(self):
        if self.thisptr != NULL:
            IDAFree(&self.thisptr)
    
    def stat(self):
        """
        Return solver statistics
        """
        cdef SolverStatistics stat = SolverStatistics()
    
        IDAGetNumSteps(self.thisptr, &stat.nsteps)
        IDAGetNumResEvals(self.thisptr, &stat.nfevals)
        IDADlsGetNumJacEvals(self.thisptr, &stat.njevals)
        IDAGetNumGEvals(self.thisptr, &stat.ngevals)
        IDAGetNumErrTestFails(self.thisptr, &stat.netfails)
        IDAGetNumNonlinSolvIters(self.thisptr, &stat.nniters)
        IDAGetNumNonlinSolvConvFails(self.thisptr, &stat.nncfails)
        
        return stat
    
    def _init_solver(self, makeConsistent=True):
        """
        Initialize the solver or restart the solver
        """
        cdef IDASettings settings
        cdef N_Vector_Serial abstolv, idv
        cdef int i, flag

        # update settings
        settings = self.settings

        settings.idv = N_Vector_Serial(2*NUMJ_MOD)
        for i in range(2*NUMJ_MOD):
            settings.idv[i] = float(1.0)

        # store initial values
        self.t = 0.0

        # Initialization
        if self.thisptr == NULL:
            self.thisptr = IDACreate()
            if self.thisptr == NULL:
                raise IDAError(IDA_MEM_NULL)

            flag = IDAInit(self.thisptr, resrob, self.t, <N_Vector>self.y.thisptr,
                           <N_Vector>self.ydot.thisptr)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)

            flag = IDASetMaxOrd(self.thisptr, settings.maxord)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)

            flag = IDADense(self.thisptr, self.y.size)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)                

            for i in range(NUMJ_MOD):
                self.tau[i] = 0.0
                
            self.userdata = TauData()

            flag = IDADlsSetDenseJacFn(self.thisptr, jacrob)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)

        # Reinitialization
        else:
            flag = IDAReInit(self.thisptr, self.t, <N_Vector>self.y.thisptr,
                             <N_Vector>self.ydot.thisptr)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)

        if settings.abstolv is None:
            flag = IDASStolerances(self.thisptr, settings.reltol, settings.abstol)
        else:
            abstolv = settings.abstolv
            flag = IDASVtolerances(self.thisptr, settings.reltol, <N_Vector>abstolv.thisptr)

        if flag != IDA_SUCCESS:
            raise IDAError(flag)

        flag = IDASetMaxNumSteps(self.thisptr, settings.mxsteps)
        if flag != IDA_SUCCESS:
            raise IDAError(flag)

        flag = IDASetMaxStep(self.thisptr, settings.hmax)
        if flag != IDA_SUCCESS:
            raise IDAError(flag)

        if not settings.idv is None:
            idv = settings.idv
            flag = IDASetId(self.thisptr, <N_Vector>idv.thisptr)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)

        flag = IDASetUserData(self.thisptr, <void*>&self.userdata)
        if flag != IDA_SUCCESS:
            raise IDAError(flag)
        
        if makeConsistent:
            flag = IDACalcIC(self.thisptr, IDA_YA_YDP_INIT, self.t + 1.0)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)
    
            flag = IDAGetConsistentIC(self.thisptr, <N_Vector>self.y.thisptr, <N_Vector>self.ydot.thisptr)
            if flag != IDA_SUCCESS:
                raise IDAError(flag)
            
            print("Consistent Initial Conditions calculated as:")
            for i in range(2*NUMJ_MOD):
                print("y[{}]={}, ydot[{}]={}".format(i, self.y[i], i, self.ydot[i]))

    cpdef calc_next_state(self, list torques):
        """Simulate the motion dynamics of the robot for the supplied
           joint torques to time plus sampling time

        Args:
            torques (list):  List of torques to be used as input for the system
                             of robot motion dynamics. Unit: Nm
        Raises:
            IDAError

        .. note:: Even if the model only describes some joints of the
            robot, the simulator expects and returns lists with
            values for all joints of the robot.
        """

        cdef int flag
        cdef realtype tret
        cdef int i

        for i in range(NUMJ_MOD):
            self.userdata.tau[i] = float(torques[SIMJOINTS[i]])

        self.t += self.sampleTime

        flag = IDASolve(self.thisptr, self.t, &tret, <N_Vector>self.y.thisptr, <N_Vector>self.ydot.thisptr, IDA_NORMAL)
        
        #print("t: {}, flag: {}, tret: {}".format(self.t, int(flag), float(tret)))
        
        #for i in range(2*NUMJ_MOD):
        #    print("tau[{}]={}, y[{}]={}".format(i, self.userdata.tau[i], i, self.y[i]))
            
        if flag < 0:
            raise IDAError(flag)

    cpdef set_state(self, list q, list dq, makeconsistent):
        """Set the current state of the simulated
        system of robot motion dynamics.
        
        Args:
            q (list): Joint positions the system should be set to

            dq (list):  Joint velocities the system should be set to

            makeconsistent (bool): Recompute consistent values for the
                                   solver. 

        .. note:: Even if the model only describes some joints of the
            robot, the simulator expects and returns lists with
            values for all joints of the robot.
        """

        for i in range(NUMJ_MOD):
            self.y[i*2] = float(q[SIMJOINTS[i]])
    
        for i in range(NUMJ_MOD):
            self.y[i*2+1] = float(dq[SIMJOINTS[i]])
            
        self._init_solver(makeconsistent)    
    
    def get_joint_positions(self):
        """Return the current joint positions from the state of the simulated
        system of robot motion dynamics.
        
        Returns:
            List of joint positions in rad

        .. note:: Even if the model only describes some joints of the
            robot, the simulator expects and returns lists with
            values for all joints of the robot.
        """

        cdef int i = 0
        q = [0.0] * NUMJ_ROB

        for i in range(NUMJ_MOD):
            q[SIMJOINTS[i]] = self.y[i*2]

        return q
    
    def get_joint_velocities(self):
        """Return the current joint velocities from the state of the simulated
        system of robot motion dynamics.
        
        Returns:
            List of joint velocities in rad/s

        .. note:: Even if the model only describes some joints of the
            robot, the simulator expects and returns lists with
            values for all joints of the robot.
        """

        cdef int i = 0
        dq = [0.0] * NUMJ_ROB

        for i in range(NUMJ_MOD):
            dq[SIMJOINTS[i]] = self.y[1+i*2]

        return dq









