'''
Template script for simulating the position or torque based controllers for the
LBR via FRI using $stiffness.strategy=10 or $stifness.strategy=30, respectively.
It works using the python scripts based on the frictl_template using the 
simulate=True flag. The default IP address is 127.0.0.1, port 49938. 
Make a copy of this file and edit to fill your needs while maintaining the same
structure.
'''

# use print as function, so it supports the end='' syntax
from __future__ import print_function
import sys
from multiprocessing import Process, Value
from ctypes import c_bool
import time

import numpy as np

from fricomm import FRI_STATE_CMD, FRI_CTRL_POSITION, FRI_CTRL_JNT_IMP, LBR_MNJ
from fricomm import FRI_QUALITY_PERFECT 
from fripy import FRIClient

class LBRSimulator(object):
    
    def __init__(self, robotModel, initSetPoint=np.zeros((LBR_MNJ,)), gravityFunct=None, debug=False):
        self._ZEROS_JNT = list(np.zeros((LBR_MNJ,)))
        
        self.debug = debug
        
        self.doStop = Value(c_bool, False)
        self.isRunning = Value(c_bool, False)
        self.process = Process(target=self._run, args=[])
        
        self._robotModel = robotModel
        self.sampleTime = robotModel.sampleTime
        self._ctrlStrategy = FRI_CTRL_JNT_IMP # FRI_CTRL_POSITION
        
        self._gravityFunct = gravityFunct


    def start(self):
        self.process.start()
    
    def stop(self):
        self.doStop.value = True
        self.process.join()
        
    def _run(self):
        doAbort = False
        
        k = 0
        sys.stdout = open("/tmp/frisim.log", "w")
        self.isRunning.value = True
        
        fri = FRIClient(debug=self.debug, verbose=self.debug)
        fri.set_control_strategy(self._ctrlStrategy)
        fri.open(0.010*100.0)
        fri.msr.data.msrJntPos = self._robotModel.get_joint_positions()
        
        while not doAbort:
            mytime = time.time()
            
            if self.debug:
                print("Quality {}, state {}".format(fri.msr.intf.quality, fri.msr.intf.state))
                   
            if (fri.msr.intf.quality is  FRI_QUALITY_PERFECT and 
                fri.msr.intf.state is not FRI_STATE_CMD):
                fri.start()
                
            if fri.msr.robot.control is FRI_CTRL_POSITION:
                fri.msr.data.msrJntPos = self._lbrsim_jntpos_compute_next_state(
                                                        fri.cmd.cmd.jntPos)
            elif fri.msr.robot.control is FRI_CTRL_JNT_IMP:
                if fri.msr.intf.state is FRI_STATE_CMD:
                    k=k+1
                    fri.msr.data.msrJntTrq = fri.cmd.cmd.addJntTrq[:]
                else:
                    fri.msr.data.msrJntTrq = self._ZEROS_JNT[:]
                    
                if self._gravityFunct is not None:
                    gravpos = np.array(fri.msr.data.msrJntPos)[self._robotModel.simulatedJoints]
                    gravtorque = self._gravityFunct(gravpos)
                    
                    curjnt = 0
                    for simjnt in self._robotModel.simulatedJoints:                            
                        fri.msr.data.msrJntTrq[simjnt] += gravtorque[curjnt]
                        curjnt += 1
                        
                if (fri.msr.intf.state is FRI_STATE_CMD):# and (k>290):
                    fri.msr.data.msrJntPos = self._jntimp_compute_next_state(fri.msr.data.msrJntTrq)
                #print('{}:  fri.msr.data.msrJntTrq={}'.format(k, fri.msr.data.msrJntTrq))
            
            if self.debug:
                print( k, ": time needed: ", time.time()-mytime)        
                    
          
            if self.doStop.value:
                print("doStop requested")                
                # Stop KRL program
                fri.msr.krl.intData[0] = 2
                fri.send_data()
                fri.close()
                doAbort = True
                break
            
            # KRL program stops
            if fri.cmd.krl.intData[1] == 2:        
                print("Stop by server requested")
                fri.msr.krl.intData[0] = 2
                fri.send_data()
                fri.close()
                doAbort = True
                break
                
            if self.debug:
                print("Send ", end="")            
                mytime = time.time()
            
            fri.send_data()
            if self.debug:
                print(k, ": time needed for send: ", time.time()-mytime)
                mytime = time.time()
                print("Recv ", end="")
            
            
            fri.recv_data()
            if fri._dont_answer:
                fri._dont_answer = False
                fri.recv_data()
                
            if self.debug:
                print(k, ": time needed for recv: ", time.time()-mytime)
            
        self.isRunning.value = False
            
    def _jntimp_compute_next_state(self, add_trq):        
        self._robotModel.calc_next_state(add_trq)        
        return self._robotModel.get_joint_positions()
